import enum
from dataclasses import dataclass
from typing import Any, Tuple, Dict, Union, Optional

from spacepackets.cfdp import TransmissionMode

# Com Interface Types
ComIfValueT = Tuple[str, Any]
ComIfDictT = Dict[str, ComIfValueT]


@dataclass
class TreeCommandingParams:
    cmd_path: Optional[str]


@dataclass
class CfdpParams:
    """Simplified dataclass to model the most important CFDP parameters. This can for example
    be used to internalize CFDP CLI or GUI options."""

    source_file: str = ""
    dest_file: str = ""
    closure_requested: bool = False
    transmission_mode: TransmissionMode = TransmissionMode.UNACKNOWLEDGED
    proxy_op: bool = False


def default_json_path() -> str:
    return "tmtc_conf.json"


class CoreComInterfaces(str, enum.Enum):
    DUMMY = "dummy"
    UDP = "udp"
    TCP = "tcp"
    SERIAL_COBS = "serial_cobs"
    SERIAL_DLE = "serial_dle"
    SERIAL_QEMU = "serial_qemu"
    UNSPECIFIED = "unspec"


CORE_COM_IF_DICT: ComIfDictT = {
    CoreComInterfaces.DUMMY: ("Dummy Interface", None),
    CoreComInterfaces.UDP: ("TCP/IP with UDP datagrams", None),
    CoreComInterfaces.TCP: ("TCP/IP with TCP", None),
    CoreComInterfaces.SERIAL_COBS: ("Serial Interace with COBS encoding", None),
    CoreComInterfaces.SERIAL_DLE: ("Serial Interace with DLE encoding", None),
    CoreComInterfaces.SERIAL_QEMU: ("Serial Interface using QEMU", None),
    CoreComInterfaces.UNSPECIFIED: ("Unspecified", None),
}


# Mode options, set by args parser
class CoreModeList(enum.IntEnum):
    """These are the core modes which will be translated to different TC and TM modes
    for the CCSDS backend

    1. ONE_QUEUE_MODE: This mode is optimized to handle one queue. It will configure the backend
       to request program termination upon finishing the queue handling. This is also the
       appropriate solution for single commands where the queue only consists of one telecommand.
    2. LISTENER_MODE: Only listen to TM
    3. MULTI_INTERACTIVE_QUEUE_MODE:
    """

    #
    ONE_QUEUE_MODE = 0
    LISTENER_MODE = 1
    # This mode is optimized for the handling of multiple queues. It will configure the backend
    # to request additional queues or a mode change from the user instead of requesting program
    # termination
    MULTI_INTERACTIVE_QUEUE_MODE = 3
    # The program will not do anything in this mode. This includes polling TM and sending any TCs
    IDLE = 5


class CoreModeConverter:
    @staticmethod
    def get_str(mode: Union[CoreModeList, int]) -> str:
        if mode == CoreModeList.LISTENER_MODE:
            return "listener"
        elif mode == CoreModeList.ONE_QUEUE_MODE:
            return "one-q"
        elif mode == CoreModeList.MULTI_INTERACTIVE_QUEUE_MODE:
            return "multi-q"
        elif mode == CoreModeList.IDLE:
            return "idle"
        else:
            return ""



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
from tmtccmd.version import get_version

# -- Project information -----------------------------------------------------

project = "tmtccmd"
copyright = "2021-2023, Robin Mueller"
author = "Robin Mueller"

# The full version, including alpha/beta/rc tags
version = release = get_version()

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.doctest",
    "sphinx_rtd_theme",
]

# Disable the doctests of the full package because those would require the explicit specification
# of imports. The doctests inside the source code are covered by pytest, using the --doctest-modules
# configuration option.
doctest_test_doctest_blocks = ""
doctest_global_setup = """
import sys
"""

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# List of directories, relative to source directory, that shouldn't be searched
# for source files.
exclude_trees = ["_build"]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "serial": ("https://pyserial.readthedocs.io/en/latest/", None),
    "cfdp-py": ("https://cfdp-py.readthedocs.io/en/latest/", None),
    "spacepackets": ("https://spacepackets.readthedocs.io/en/latest/", None),
    "prompt-toolkit": ("https://python-prompt-toolkit.readthedocs.io/en/latest/", None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "../misc/logo_medium.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# -- Options for LaTeX output --------------------------------------------------

# The name of an image file (relative to this directory) to place at the top of
# the title page.
latex_logo = "../misc/logo.png"



"""Core components for mode commanding (custom PUS service)."""

import enum
import struct
from typing import Union

from deprecated.sphinx import deprecated
from spacepackets.ecss import PusTelecommand
from tmtccmd.pus import CustomFsfwPusService
from tmtccmd.pus.s200_fsfw_mode import Subservice


class Mode(enum.IntEnum):
    """Standard modes when commanding objects. These mode IDs are reserved by the FSFW,
    so it is recommended to avoid these numbers for custom modes."""

    OFF = 0
    ON = 1
    NORMAL = 2
    RAW = 3


def pack_mode_data(object_id: bytes, mode: Union[Mode, int], submode: int) -> bytearray:
    """FSFW modes: Mode 0: Off, Mode 1: Mode On, Mode 2: Mode Normal, Mode 3: Mode Raw"""
    mode_data = bytearray()
    mode_data += object_id + struct.pack("!I", mode) + struct.pack("B", submode)
    return mode_data


@deprecated(
    version="v4.0.0a2",
    reason="use create... API instead",
)
def pack_mode_command(
    object_id: bytes, mode: Union[int, Mode], submode: int
) -> PusTelecommand:
    return create_mode_command(object_id, mode, submode)


def create_mode_command(
    object_id: bytes, mode: Union[int, Mode], submode: int
) -> PusTelecommand:
    return PusTelecommand(
        service=CustomFsfwPusService.SERVICE_200_MODE,
        subservice=Subservice.TC_MODE_COMMAND,
        app_data=pack_mode_data(object_id, mode, submode),
    )


def create_read_mode_command(object_id: bytes) -> PusTelecommand:
    return PusTelecommand(
        service=CustomFsfwPusService.SERVICE_200_MODE,
        subservice=Subservice.TC_MODE_READ,
        app_data=object_id,
    )


def create_announce_mode_command(object_id: bytes) -> PusTelecommand:
    return PusTelecommand(
        service=CustomFsfwPusService.SERVICE_200_MODE,
        subservice=Subservice.TC_MODE_ANNOUNCE,
        app_data=object_id,
    )


def create_announce_mode_recursive_command(object_id: bytes) -> PusTelecommand:
    return PusTelecommand(
        service=CustomFsfwPusService.SERVICE_200_MODE,
        subservice=Subservice.TC_MODE_ANNOUNCE_RECURSIVE,
        app_data=object_id,
    )



import warnings
from tmtccmd.fsfw.tmtc_printer import FsfwTmTcPrinter  # noqa: F401


warnings.warn(
    "This module is deprecated, use tmtccmd.fsfw.tmtc_printer",
    DeprecationWarning,
    stacklevel=2,
)



"""Contains core methods called by entry point files to setup and start a tmtccmd application.
It also re-exports commonly used classes and functions.
"""

import logging
import sys
import os
from datetime import timedelta
from typing import Union, cast, Optional

from tmtccmd.config.args import ProcedureParamsWrapper
from tmtccmd.core.ccsds_backend import CcsdsTmtcBackend
from tmtccmd.core.base import FrontendBase, BackendRequest
from tmtccmd.tmtc.ccsds_tm_listener import CcsdsTmListener
from tmtccmd.config import (
    HookBase,
    backend_mode_conversion,
    SetupWrapper,
    SetupParams,
    PreArgsParsingWrapper,
    CoreModeConverter,
    CoreModeList,
    TreeCommandingParams,
)
from tmtccmd.core.ccsds_backend import BackendBase
from tmtccmd.tmtc import TmTypes, TmHandlerBase, CcsdsTmHandler
from tmtccmd.core import ModeWrapper
from tmtccmd.tmtc import (
    TreeCommandingProcedure,
    TcProcedureBase,
    ProcedureWrapper,
    TcHandlerBase,
)


__TMTCCMD_LOGGER = logging.getLogger(__name__)

__SETUP_WAS_CALLED = False
__SETUP_FOR_GUI = False


def init_logger(propagate: bool = False, log_level: int = logging.INFO):
    """Initiate the library internal logger. There are various ways how to use the logging support
    of tmtccmd in an application.

    1. Usage of a custom application logger, possibly a root logger created with
       :py:func:`logging.getLogger`. In that case, this function does not have to be called if
       the goal is to have the library logs be emitted by the custom logger. It might still make
       sense to apply the library console logging format to the application logger using
       :py:func:`tmtccmd.logging.add_colorlog_console_logger`.
    2. Usage of a custom application logger, but the library logs should not be emitted
       by that logger. In that case, the propagation can be disabled but this function can be
       used to still set up the library logger.
    3. No usage of logging in the application but the logs of the library should still be emitted.
       In that case, this function should still be called to set the log level and set up
       formatting. The propagation flag does not matter and can be left at the default level.

    :param propagate: Specifies whether logs are propagated. If the user wants to use an own (root)
        logger and does not wish the logs of the library to be propagated to that logger,
        this should be set to False, which is the default.
    :param log_level: Sets the log level of the library logger
    """
    from tmtccmd.logging import __setup_tmtc_console_logger

    __setup_tmtc_console_logger(__TMTCCMD_LOGGER, propagate, log_level)


def get_lib_logger() -> logging.Logger:
    """Get the library logger. Please note that this logger can be configured by
    calling :py:func:`init_logger`. This might make sense depending on how the library logger
    is used."""
    return __TMTCCMD_LOGGER


def setup(setup_args: SetupWrapper):
    """This function needs to be called first before running the TMTC commander core. The setup
    arguments encapsulate all required arguments for the TMTC commander.

    :param setup_args:     Setup arguments
    """
    global __SETUP_WAS_CALLED, __SETUP_FOR_GUI

    if os.name == "nt":
        import colorama

        colorama.init()
    __SETUP_FOR_GUI = setup_args.params.use_gui
    __SETUP_WAS_CALLED = True


def start(
    tmtc_backend: BackendBase,
    hook_obj: HookBase,
    tmtc_frontend: Optional[FrontendBase] = None,
    app_name: str = "TMTC Commander",
):
    """This is the primary function to run the TMTC commander. Users should call this function to
    start the TMTC commander. Please note that :py:func:`setup` needs to be
    called before this function.  You also need to build a TMTC backend
    instance and pass it to this call. You can use :py:func:`create_default_tmtc_backend`
    to create a generic backend.

    :param tmtc_backend:        Custom backend can be passed here. Otherwise, a default backend
                                will be created
    :param hook_obj:
    :param tmtc_frontend:       Custom frontend can be passed here. Otherwise, a default frontend
                                will be created
    :param app_name:            Name of application. Will be displayed in GUI
    :raises RunTimeError:  if :py:func:`setup` was not called before
    :return:
    """
    global __SETUP_WAS_CALLED, __SETUP_FOR_GUI
    if not __SETUP_WAS_CALLED:
        __TMTCCMD_LOGGER.warning("setup_tmtccmd was not called first. Call it first")
        sys.exit(1)
    if __SETUP_FOR_GUI:
        __start_tmtc_commander_qt_gui(
            tmtc_frontend=tmtc_frontend,
            hook_obj=hook_obj,
            tmtc_backend=tmtc_backend,
            app_name=app_name,
        )
    else:
        __start_tmtc_commander_cli(tmtc_backend=tmtc_backend)


def init_printout(use_gui: bool):
    from tmtccmd.version import get_version

    if use_gui:
        print(f"-- tmtccmd v{get_version()} GUI Mode --")
    else:
        print(f"-- tmtccmd v{get_version()} CLI Mode --")


def __start_tmtc_commander_cli(tmtc_backend: BackendBase):
    tmtc_backend.open_com_if()


def __start_tmtc_commander_qt_gui(
    tmtc_backend: BackendBase,
    hook_obj: HookBase,
    tmtc_frontend: Union[None, FrontendBase] = None,
    app_name: str = "TMTC Commander",
):
    global __SETUP_WAS_CALLED
    try:
        from PyQt6.QtWidgets import QApplication

        if not __SETUP_WAS_CALLED:
            __TMTCCMD_LOGGER.warning(
                "setup_tmtccmd was not called first. Call it first"
            )
            sys.exit(1)
        app = QApplication([app_name])
        if tmtc_frontend is None:
            from tmtccmd.gui import TmTcFrontend
            from tmtccmd.core.ccsds_backend import CcsdsTmtcBackend

            tmtc_frontend = TmTcFrontend(
                hook_obj=hook_obj,
                tmtc_backend=cast(CcsdsTmtcBackend, tmtc_backend),
                app_name=app_name,
            )
        tmtc_frontend.start(app)
    except ImportError as e:
        __TMTCCMD_LOGGER.exception(e)
        sys.exit(1)


def create_default_tmtc_backend(
    setup_wrapper: SetupWrapper,
    tm_handler: TmHandlerBase,
    tc_handler: TcHandlerBase,
    init_procedure: Optional[ProcedureWrapper],
) -> BackendBase:
    """Creates a default TMTC backend instance which can be passed to the tmtccmd runner

    :param init_procedure:
    :param tc_handler:
    :param setup_wrapper:
    :param tm_handler:
    :return:
    """
    global __SETUP_WAS_CALLED

    from typing import cast

    if not __SETUP_WAS_CALLED:
        __TMTCCMD_LOGGER.warning("setup_tmtccmd was not called first. Call it first")
        sys.exit(1)
    if tm_handler is None:
        __TMTCCMD_LOGGER.warning(
            "No TM Handler specified! Make sure to specify at least one TM handler"
        )
        sys.exit(1)
    else:
        if tm_handler.get_type() == TmTypes.CCSDS_SPACE_PACKETS:
            tm_handler = cast(CcsdsTmHandler, tm_handler)
    com_if = setup_wrapper.params.com_if
    if com_if is None:
        com_if = setup_wrapper.hook_obj.get_communication_interface(
            setup_wrapper.params.com_if_id
        )
    tm_listener = CcsdsTmListener(tm_handler)
    mode_wrapper = ModeWrapper()
    backend_mode_conversion(setup_wrapper.params.mode, mode_wrapper)
    if setup_wrapper.params.mode == CoreModeConverter.get_str(
        CoreModeList.LISTENER_MODE
    ):
        print("-- Backend Listener Mode --")
    elif setup_wrapper.params.mode == CoreModeConverter.get_str(
        CoreModeList.ONE_QUEUE_MODE
    ):
        print("-- One Queue Mode --")
    elif setup_wrapper.params.mode == CoreModeConverter.get_str(
        CoreModeList.MULTI_INTERACTIVE_QUEUE_MODE
    ):
        print("-- Multi Queue Mode --")
    assert com_if is not None
    # The global variables are set by the argument parser.
    tmtc_backend = CcsdsTmtcBackend(
        com_if=com_if,
        tm_listener=tm_listener,
        tc_handler=tc_handler,
        tc_mode=mode_wrapper.tc_mode,
        tm_mode=mode_wrapper.tm_mode,
    )
    if setup_wrapper.params.backend_params.listener:
        tmtc_backend.keep_listener_mode = True
    tmtc_backend.inter_cmd_delay = timedelta(
        seconds=setup_wrapper.params.cmd_params.delay
    )
    if init_procedure is not None:
        tmtc_backend.current_procedure = init_procedure.procedure
    return tmtc_backend


def setup_backend_def_procedure(
    backend: CcsdsTmtcBackend, tmtc_params: TreeCommandingProcedure
):
    assert tmtc_params.cmd_path is not None
    backend.current_procedure = TreeCommandingProcedure(tmtc_params.cmd_path)



from __future__ import annotations


class FileInfo:
    def __init__(
        self,
        file_name: str,
        repo_path: str,
        file_size: int = 0,
        lock_status: bool = False,
    ):
        self.file_name = file_name
        self.repo_path = repo_path
        self.file_size = file_size
        self.lock_status = lock_status



from spacepackets.countdown import Countdown, time_ms  # noqa: F401

import warnings

warnings.warn(
    "the countdown module is deprecated and was moved to spacepackets.countdown",
    DeprecationWarning,
    stacklevel=2,
)



import argparse
from unittest import TestCase

from tmtccmd.config.args import (
    add_tmtc_mode_arguments,
    add_generic_arguments,
    add_default_com_if_arguments,
    add_tree_commanding_arguments,
)


class TestArgsParsing(TestCase):
    def setUp(self) -> None:
        self.arg_parser = argparse.ArgumentParser()

    def test_basic_arg_parser_mode_empty(self):
        add_tmtc_mode_arguments(self.arg_parser)
        args = self.arg_parser.parse_args([])
        self.assertEqual(args.mode, None)

    def test_valid_argument_0(self):
        add_tmtc_mode_arguments(self.arg_parser)
        args = self.arg_parser.parse_args(["-m", "one-q"])
        self.assertEqual(args.mode, "one-q")

    def test_valid_argument_1(self):
        add_tmtc_mode_arguments(self.arg_parser)
        args = self.arg_parser.parse_args(["--mode", "one-q"])
        self.assertEqual(args.mode, "one-q")

    def test_def_proc_argument_empty(self):
        add_tree_commanding_arguments(self.arg_parser)
        args = self.arg_parser.parse_args([])
        self.assertIsNone(args.cmd_path)

    def test_def_proc_argument_valid(self):
        add_tree_commanding_arguments(self.arg_parser)
        args = self.arg_parser.parse_args(["-p", "/PING"])
        self.assertEqual(args.cmd_path, "/PING")
        self.assertEqual(args.print_tree, None)

    def test_generic_arguments_empty(self):
        add_generic_arguments(self.arg_parser)
        args = self.arg_parser.parse_args([])
        self.assertIsNone(args.delay)
        self.assertFalse(args.gui)

    def test_generic_arguments_valid(self):
        add_generic_arguments(self.arg_parser)
        args = self.arg_parser.parse_args(["-d", "2.0", "-g"])
        self.assertEqual(args.delay, 2.0)
        self.assertEqual(args.gui, True)

    def test_com_if_arguments_empty(self):
        add_default_com_if_arguments(self.arg_parser)
        args = self.arg_parser.parse_args([])
        self.assertEqual(args.com_if, "unspec")

    def test_com_if_arguments_valid(self):
        add_default_com_if_arguments(self.arg_parser)
        args = self.arg_parser.parse_args(["-c", "udp"])
        self.assertEqual(args.com_if, "udp")

    def test_tree_print_args_0(self):
        add_tree_commanding_arguments(self.arg_parser)
        args = self.arg_parser.parse_args(["-T"])
        self.assertEqual(args.print_tree, [])
        args = self.arg_parser.parse_args(["--print-tree"])
        self.assertEqual(args.print_tree, [])
        args = self.arg_parser.parse_args(["--pt"])
        self.assertEqual(args.print_tree, [])

    def test_tree_print_args_1(self):
        add_tree_commanding_arguments(self.arg_parser)
        args = self.arg_parser.parse_args(["-T", "b", "2"])
        self.assertEqual(args.print_tree, ["b", "2"])



from __future__ import annotations

from spacepackets import SpacePacketHeader
from spacepackets.ccsds.spacepacket import PacketId, PacketSeqCtrl
from spacepackets.ecss import (
    Ptc,
    PusTelemetry,
)
from spacepackets.ecss.defs import PusService
from spacepackets.ecss.tm import AbstractPusTm

from tmtccmd.pus.s20_fsfw_param import (
    CustomSubservice,
)
from tmtccmd.pus.s20_fsfw_param_defs import ParameterId, Parameter, FsfwParamId


class Service20ParamDumpWrapper:
    def __init__(self, param_tm: Service20FsfwTm):
        self.param_tm = param_tm
        if self.param_tm.subservice != CustomSubservice.TM_DUMP_REPLY:
            raise ValueError(f"subservice is not {CustomSubservice.TM_DUMP_REPLY}")

    @property
    def base_tm(self) -> PusTelemetry:
        return self.param_tm.pus_tm

    def get_param(self) -> Parameter:
        """Tries to build a :py:class:`Parameter` from the own raw telemetry data.

        :raises ValueError: Telemetry source data too short.
        """
        if len(self.param_tm.source_data) < 12:
            raise ValueError("source data too short to hold parameter")
        param_id = ParameterId.unpack(self.param_tm.source_data[4:8])
        try:
            ptc = Ptc(self.param_tm.source_data[8])
        except TypeError:
            raise ValueError(f"unknown PTC {self.param_tm.source_data[8]}")
        pfc = self.param_tm.source_data[9]
        rows = self.param_tm.source_data[10]
        columns = self.param_tm.source_data[11]
        param_data = self.param_tm.source_data[12:]
        return Parameter(
            fsfw_param_id=FsfwParamId(
                object_id=self.param_tm.source_data[0:4],
                param_id=param_id,
                ptc=Ptc(ptc),
                pfc=pfc,
                rows=rows,
                columns=columns,
            ),
            param_raw=param_data,
        )


class Service20FsfwTm(AbstractPusTm):
    def __init__(
        self,
        subservice: int,
        source_data: bytes,
        timestamp: bytes,
        apid: int = 0,
    ):
        self.pus_tm = PusTelemetry(
            service=PusService.S20_PARAMETER,
            subservice=subservice,
            source_data=source_data,
            timestamp=timestamp,
            apid=apid,
        )

    @staticmethod
    def __common_checks(tm: PusTelemetry):
        if tm.service != 20:
            raise ValueError("service ID is not 20")
        if len(tm.source_data) < 4:
            raise ValueError(
                "source data must include object ID (4 bytes) as minimal data"
            )

    @classmethod
    def unpack(cls, raw_telemetry: bytes, timestamp_len: int) -> Service20FsfwTm:
        instance = cls.empty()
        instance.pus_tm = PusTelemetry.unpack(
            data=raw_telemetry, timestamp_len=timestamp_len
        )
        Service20FsfwTm.__common_checks(instance.pus_tm)
        return instance

    @classmethod
    def from_tm(cls, tm: PusTelemetry):
        instance = cls.empty()
        instance.pus_tm = tm
        Service20FsfwTm.__common_checks(instance.pus_tm)
        return instance

    def pack(self) -> bytearray:
        return self.pus_tm.pack()

    @property
    def timestamp(self) -> bytes:
        return self.pus_tm.timestamp

    @property
    def object_id(self) -> bytes:
        return bytes(self.source_data[0:4])

    @property
    def service(self) -> int:
        return self.pus_tm.service

    @property
    def subservice(self) -> int:
        return self.pus_tm.subservice

    @property
    def source_data(self) -> bytes:
        return self.pus_tm.source_data

    @classmethod
    def empty(cls) -> Service20FsfwTm:
        return cls(
            subservice=0,
            apid=0,
            source_data=bytes([0, 0, 0, 0]),
            timestamp=bytes(),
        )

    @property
    def sp_header(self) -> SpacePacketHeader:
        return self.pus_tm.space_packet_header

    @property
    def ccsds_version(self) -> int:
        return self.pus_tm.ccsds_version

    @property
    def packet_id(self) -> PacketId:
        return self.pus_tm.packet_id

    @property
    def packet_seq_control(self) -> PacketSeqCtrl:
        return self.pus_tm.packet_seq_control

    def __eq__(self, other: object):
        if not isinstance(other, Service20FsfwTm):
            return False
        return self.pus_tm == other.pus_tm



from unittest import TestCase

from spacepackets.ecss import PusTelecommand
from tmtccmd.com.dummy import DummyComIF


class TestDummy(TestCase):
    def test_dummy_if(self):
        dummy_com_if = DummyComIF()
        self.assertFalse(dummy_com_if.is_open())
        dummy_com_if.open()
        self.assertTrue(dummy_com_if.is_open())
        self.assertFalse(dummy_com_if.initialized)
        dummy_com_if.initialize()
        self.assertTrue(dummy_com_if.initialized)
        self.assertFalse(dummy_com_if.data_available())
        dummy_com_if.send(PusTelecommand(apid=0x02, service=17, subservice=1).pack())
        self.assertTrue(dummy_com_if.data_available())
        replies = dummy_com_if.receive()
        # Full verification set (acceptance, start and completion) and ping reply
        self.assertEqual(len(replies), 4)



import time
import select
import socket
from typing import Any
from unittest import TestCase

from tmtccmd.com.tcpip_utils import EthAddr
from tmtccmd.com.udp import UdpClient

LOCALHOST = "127.0.0.1"


class TestUdpIf(TestCase):
    def setUp(self) -> None:
        self.udp_server = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.addr = (LOCALHOST, 7777)
        self.udp_server.bind(self.addr)
        self.udp_client = UdpClient("udp", send_address=EthAddr.from_tuple(self.addr))
        self.udp_client.initialize()

    def test_basic(self):
        self.assertEqual(self.udp_client.id, "udp")
        self._open()

    def test_send(self):
        self._open()
        self._simple_send(bytes([0, 1, 2, 3]))

    def test_recv(self):
        self._open()
        data = bytes([0, 1, 2, 3])
        sender_addr = self._simple_send(data)
        self.udp_server.sendto(data, sender_addr)
        time.sleep(0.05)
        self.assertTrue(self.udp_client.data_available())
        data_recv = self.udp_client.receive()
        self.assertEqual(len(data_recv), 1)
        self.assertEqual(data_recv[0], data)

    def _simple_send(self, data: bytes) -> Any:
        self.udp_client.send(data)
        ready = select.select([self.udp_server], [], [], 0.1)
        self.assertTrue(ready[0])
        data_recv, sender_addr = self.udp_server.recvfrom(4096)
        self.assertEqual(data, data_recv)
        return sender_addr

    def _open(self):
        self.udp_client.open()
        self.assertTrue(self.udp_client.is_open())

    def tearDown(self) -> None:
        self.udp_client.close()
        self.udp_server.close()



"""Contains definitions and functions related to PUS Service 3 Telecommands.
"""

import struct
from typing import Tuple

from deprecated.sphinx import deprecated
from spacepackets.ecss.tc import PusTelecommand
from spacepackets.ecss.pus_3_hk import Subservice


def make_sid(object_id: bytes, set_id: int) -> bytearray:
    sid_raw = bytearray()
    sid_raw += object_id + struct.pack(">I", set_id)
    return sid_raw


def make_interval(interval_seconds: float) -> bytearray:
    return bytearray(struct.pack("!f", interval_seconds))


@deprecated(
    version="v4.0.0a2",
    reason="use create... API instead",
)
def enable_periodic_hk_command(diag: bool, sid: bytes) -> PusTelecommand:
    return create_enable_periodic_hk_command_with_diag(diag, sid)


def create_enable_periodic_hk_command(sid: bytes) -> PusTelecommand:
    return __generate_periodic_hk_command(enable=True, sid=sid)


@deprecated(
    version="v6.0.0rc0",
    reason="use diagnostic agnostic API if possible",
)
def create_enable_periodic_hk_command_with_diag(
    diag: bool, sid: bytes
) -> PusTelecommand:
    return __generate_periodic_hk_command_legacy(diag=diag, enable=True, sid=sid)


@deprecated(
    version="v6.0.0rc0",
    reason="use diagnostic agnostic API if possible",
)
def create_enable_periodic_hk_command_with_interval_with_diag(
    diag: bool, sid: bytes, interval_seconds: float
) -> Tuple[PusTelecommand, PusTelecommand]:
    cmd0 = create_modify_collection_interval_cmd_with_diag(diag, sid, interval_seconds)
    cmd1 = __generate_periodic_hk_command_legacy(diag=diag, enable=True, sid=sid)
    return cmd0, cmd1


def create_enable_periodic_hk_command_with_interval(
    sid: bytes, interval_seconds: float
) -> Tuple[PusTelecommand, PusTelecommand]:
    cmd0 = create_modify_collection_interval_cmd(sid, interval_seconds)
    cmd1 = __generate_periodic_hk_command(enable=True, sid=sid)
    return cmd0, cmd1


@deprecated(
    version="v4.0.0a2",
    reason="use create... API instead",
)
def enable_periodic_hk_command_with_interval(
    diag: bool, sid: bytes, interval_seconds: float
) -> Tuple[PusTelecommand, PusTelecommand]:
    return create_enable_periodic_hk_command_with_interval_with_diag(
        diag, sid, interval_seconds
    )


def create_disable_periodic_hk_command(sid: bytes) -> PusTelecommand:
    return __generate_periodic_hk_command(enable=False, sid=sid)


@deprecated(
    version="v6.0.0rc0",
    reason="use diagnostic agnostic API if possible",
)
def create_disable_periodic_hk_command_with_diag(
    diag: bool, sid: bytes
) -> PusTelecommand:
    return __generate_periodic_hk_command_legacy(diag=diag, enable=False, sid=sid)


@deprecated(
    version="v4.0.0a2",
    reason="use create... API instead",
)
def disable_periodic_hk_command(diag: bool, sid: bytes) -> PusTelecommand:
    return create_disable_periodic_hk_command_with_diag(diag, sid)


def __generate_periodic_hk_command(enable: bool, sid: bytes) -> PusTelecommand:
    app_data = bytearray(sid)
    if enable:
        subservice = Subservice.TC_ENABLE_PERIODIC_HK_GEN
    else:
        subservice = Subservice.TC_DISABLE_PERIODIC_HK_GEN
    return PusTelecommand(service=3, subservice=subservice, app_data=app_data)


def __generate_periodic_hk_command_legacy(
    diag: bool, enable: bool, sid: bytes
) -> PusTelecommand:
    app_data = bytearray(sid)
    if enable:
        if diag:
            subservice = Subservice.TC_ENABLE_PERIODIC_DIAGNOSTICS_GEN
        else:
            subservice = Subservice.TC_ENABLE_PERIODIC_HK_GEN
    else:
        if diag:
            subservice = Subservice.TC_DISABLE_PERIODIC_DIAGNOSTICS_GEN
        else:
            subservice = Subservice.TC_DISABLE_PERIODIC_HK_GEN
    return PusTelecommand(service=3, subservice=subservice, app_data=app_data)


@deprecated(
    version="v4.0.0a2",
    reason="use create... API instead",
)
def modify_collection_interval(
    diag: bool, sid: bytes, interval_seconds: float
) -> PusTelecommand:
    return create_modify_collection_interval_cmd_with_diag(diag, sid, interval_seconds)


def create_modify_collection_interval_cmd(
    sid: bytes, interval_seconds: float
) -> PusTelecommand:
    app_data = bytearray(sid)
    app_data += make_interval(interval_seconds)
    subservice = Subservice.TC_MODIFY_PARAMETER_REPORT_COLLECTION_INTERVAL
    return PusTelecommand(service=3, subservice=subservice, app_data=app_data)


@deprecated(
    version="v6.0.0rc0",
    reason="use diagnostic agnostic API if possible",
)
def create_modify_collection_interval_cmd_with_diag(
    diag: bool, sid: bytes, interval_seconds: float
) -> PusTelecommand:
    app_data = bytearray(sid)
    app_data += make_interval(interval_seconds)
    if diag:
        subservice = Subservice.TC_MODIFY_DIAGNOSTICS_REPORT_COLLECTION_INTERVAL
    else:
        subservice = Subservice.TC_MODIFY_PARAMETER_REPORT_COLLECTION_INTERVAL
    return PusTelecommand(service=3, subservice=subservice, app_data=app_data)


@deprecated(
    version="v4.0.0a2",
    reason="use create... API instead",
)
def generate_one_hk_command(sid: bytes) -> PusTelecommand:
    return create_request_one_hk_command(sid)


def create_request_one_hk_command(sid: bytes) -> PusTelecommand:
    return PusTelecommand(
        service=3,
        subservice=Subservice.TC_GENERATE_ONE_PARAMETER_REPORT,
        app_data=sid,
    )


@deprecated(
    version="v4.0.0a2",
    reason="use create... API instead",
)
def generate_one_diag_command(sid: bytes) -> PusTelecommand:
    return create_request_one_diag_command(sid)


def create_request_one_diag_command(sid: bytes) -> PusTelecommand:
    return PusTelecommand(
        service=3,
        subservice=Subservice.TC_GENERATE_ONE_DIAGNOSTICS_REPORT,
        app_data=sid,
    )



import csv
import os
import copy
from typing import Optional, List
from tmtccmd.util.obj_id import ObjectIdU32, ObjectIdDictT
from tmtccmd.pus.s5_fsfw_event import EventInfo, EventDictT
from tmtccmd.util.retval import RetvalDictT, RetvalInfo


def parse_fsfw_objects_csv(csv_file: str) -> Optional[ObjectIdDictT]:
    if os.path.exists(csv_file):
        obj_id_dict = dict()
        obj_id = ObjectIdU32(obj_id=0)
        with open(csv_file) as csvfile:
            csv_reader = csv.reader(csvfile, delimiter=";")
            for row in csv_reader:
                # Parse hex string
                obj_id.obj_id = int(row[0], 16)
                obj_id.name = row[1]
                obj_id_dict.update({obj_id.as_bytes: copy.copy(obj_id)})
        return obj_id_dict
    else:
        return None


def parse_fsfw_events_csv(csv_file: str) -> Optional[EventDictT]:
    if os.path.exists(csv_file):
        event_dict = dict()
        with open(csv_file) as csvfile:
            csv_reader = csv.reader(csvfile, delimiter=";")
            info = EventInfo()
            for idx, row in enumerate(csv_reader):
                if idx == 0:
                    continue
                info.id = int(row[0])
                info.name = row[2]
                info.severity = row[3]
                info.info = row[4]
                info.file_location = row[5]
                event_dict.update({info.id: copy.copy(info)})
        return event_dict
    else:
        return None


def parse_fsfw_returnvalues_csv(csv_file: str) -> Optional[RetvalDictT]:
    if os.path.exists(csv_file):
        retval_dict = dict()
        with open(csv_file) as csvfile:
            csv_reader = csv.reader(csvfile, delimiter=";")
            info = RetvalInfo()
            for idx, row in enumerate(csv_reader):
                if idx == 0:
                    continue
                # Parse hex
                id_col = str(row[0])
                if "0x" in id_col:
                    info.id = int(id_col, 0)
                else:
                    info.id = int(id_col, 16)
                info.name = row[1]
                info.info = row[2]
                info.if_name = row[5]
                retval_dict.update({info.id: copy.copy(info)})
        return retval_dict
    else:
        return None


def bit_extractor(byte: int, position: int):
    """

    :param byte:
    :param position:
    :return:
    """
    shift_number = position + (6 - 2 * (position - 1))
    return (byte >> shift_number) & 1


def validity_buffer_list(validity_buffer: bytes, num_vars: int) -> List[bool]:
    """
    :param validity_buffer: Validity buffer in bytes format
    :param num_vars: Number of variables
    :return:
    """
    valid_list = []
    counter = 0
    for index, byte in enumerate(validity_buffer):
        for bit in range(1, 9):
            if bit_extractor(byte, bit) == 1:
                valid_list.append(True)
            else:
                valid_list.append(False)
            counter += 1
            if counter == num_vars:
                return valid_list
    return valid_list



from __future__ import annotations
import dataclasses
import enum
import struct


class EventSeverity(enum.IntEnum):
    INFO = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3


@dataclasses.dataclass
class EventU32:
    severity: EventSeverity
    group_id: int
    unique_id: int

    @classmethod
    def unpack(cls, data: bytes) -> EventU32:
        if len(data) < 4:
            raise ValueError("passed data too short")
        event_raw = struct.unpack("!I", data[0:4])[0]
        return cls(
            severity=EventSeverity((event_raw >> 30) & 0b11),
            group_id=(event_raw >> 16) & 0x3FFF,
            unique_id=event_raw & 0xFFFF,
        )



#!/usr/bin/env python3
"""Example application for the TMTC Commander"""
import logging
import sys
import time
from typing import Any, Optional

from prompt_toolkit.history import FileHistory, History
from spacepackets.ccsds import CdsShortTimestamp
from spacepackets.ecss import PusTelecommand, PusTelemetry, PusVerificator
from spacepackets.ecss.pus_1_verification import Service1Tm, UnpackParams
from spacepackets.ecss.pus_17_test import Service17Tm
from spacepackets.util import UnsignedByteField

import tmtccmd
from tmtccmd import BackendRequest, CcsdsTmtcBackend, ProcedureParamsWrapper
from tmtccmd.com import ComInterface
from tmtccmd.config import (
    CmdTreeNode,
    HookBase,
    PreArgsParsingWrapper,
    SetupParams,
    SetupWrapper,
    default_json_path,
    params_to_procedure_conversion,
)
from tmtccmd.config.args import perform_tree_printout
from tmtccmd.fsfw.tmtc_printer import FsfwTmTcPrinter
from tmtccmd.logging import add_colorlog_console_logger
from tmtccmd.logging.pus import (
    RegularTmtcLogWrapper,
)
from tmtccmd.pus import VerificationWrapper
from tmtccmd.pus.s5_fsfw_event import Service5Tm
from tmtccmd.tmtc import (
    CcsdsTmHandler,
    DefaultPusQueueHelper,
    FeedWrapper,
    ProcedureWrapper,
    QueueWrapper,
    SendCbParams,
    SpecificApidHandlerBase,
    TcHandlerBase,
    TcProcedureType,
    TcQueueEntryType,
)
from tmtccmd.util import ObjectIdDictT
from spacepackets.seqcount import FileSeqCountProvider, PusFileSeqCountProvider

_LOGGER = logging.getLogger()

EXAMPLE_PUS_APID = 0xEF
EXAMPLE_CFDP_APID = 0xF0

CFDP_LOCAL_ENTITY_ID = UnsignedByteField(byte_len=2, val=1)
CFDP_REMOTE_ENTITY_ID = UnsignedByteField(byte_len=2, val=EXAMPLE_CFDP_APID)


class ExampleHookClass(HookBase):
    def __init__(self, json_cfg_path: str):
        super().__init__(json_cfg_path=json_cfg_path)

    def get_communication_interface(self, com_if_key: str) -> Optional[ComInterface]:
        assert self.cfg_path is not None
        print("Communication interface assignment function was called")
        from tmtccmd.config.com import (
            create_com_interface_cfg_default,
            create_com_interface_default,
        )

        assert self.cfg_path is not None
        cfg = create_com_interface_cfg_default(
            com_if_key=com_if_key,
            json_cfg_path=self.cfg_path,
            space_packet_ids=None,
        )
        assert cfg is not None
        return create_com_interface_default(cfg)

    def get_command_definitions(self) -> CmdTreeNode:
        root_node = CmdTreeNode.root_node()
        root_node.add_child(CmdTreeNode("ping", "Send PUS ping command"))
        root_node.add_child(CmdTreeNode("test", "Test Node"))
        root_node.children["test"].add_child(
            CmdTreeNode("event", "Send PUS event test command")
        )
        return self.build_more_complex_tree(root_node)

    def build_more_complex_tree(self, root_node: CmdTreeNode) -> CmdTreeNode:
        root_node.add_child(CmdTreeNode("system", "System Commands"))
        root_node.add_child(CmdTreeNode("acs", "ACS Subsystem"))
        root_node["acs"].add_child(CmdTreeNode("acs_ctrl", "ACS Controller"))
        root_node["acs"].add_child(CmdTreeNode("mgt", "Magnetorquer"))
        root_node["acs"]["mgt"].add_child(CmdTreeNode("set_dipoles", "Set MGT Dipoles"))
        root_node["acs"].add_child(CmdTreeNode("mgm0", "Magnetometer 0"))
        root_node["acs"].add_child(CmdTreeNode("mgm1", "Magnetometer 1"))
        mgm_node = CmdTreeNode("other cmds", "Other MGM commands")
        root_node["acs"]["mgm0"].add_child(mgm_node)
        root_node["acs"]["mgm1"].add_child(mgm_node)
        root_node.add_child(CmdTreeNode("tcs", "TCS Subsystem"))
        root_node["tcs"].add_child(CmdTreeNode("tcs_ctrl", "TCS Controller"))
        root_node["tcs"].add_child(CmdTreeNode("pt1000", "Temperature Sensor"))
        root_node["tcs"].add_child(CmdTreeNode("heater", "Heater"))
        root_node.add_child(CmdTreeNode("com", "COM Subsystem"))
        root_node["com"].add_child(CmdTreeNode("uhf_transceiver", "UHF Transceiver"))
        root_node.add_child(CmdTreeNode("eps", "EPS Subsystem"))
        root_node["eps"].add_child(CmdTreeNode("pcdu", "PCDU"))
        root_node["eps"]["pcdu"].add_child(CmdTreeNode("channel_0_on", "Channel 0 on"))
        root_node["eps"]["pcdu"].add_child(
            CmdTreeNode("channel_0_off", "Channel 0 off")
        )
        root_node["eps"]["pcdu"].add_child(CmdTreeNode("channel_1_on", "Channel 1 on"))
        root_node["eps"]["pcdu"].add_child(
            CmdTreeNode("channel_1_off", "Channel 1 off")
        )
        return root_node

    def get_cmd_history(self) -> Optional[History]:
        """Optionlly return a history class for the past command paths which will be used
        when prompting a commad path from the user in CLI mode."""
        return FileHistory(".tmtc-cli-history.txt")

    def perform_mode_operation(self, _tmtc_backend: CcsdsTmtcBackend, _mode: int):
        _LOGGER.info("Mode operation hook was called")
        pass

    def get_object_ids(self) -> ObjectIdDictT:
        from tmtccmd.config.objects import get_core_object_ids

        return get_core_object_ids()


class PusTmHandler(SpecificApidHandlerBase):
    def __init__(
        self,
        verif_wrapper: VerificationWrapper,
        printer: FsfwTmTcPrinter,
    ):
        super().__init__(EXAMPLE_PUS_APID, None)
        self.printer = printer
        self.verif_wrapper = verif_wrapper
        assert self.printer.file_logger is not None

    def handle_tm(self, packet: bytes, _user_args: Any):
        try:
            tm_packet = PusTelemetry.unpack(
                packet, timestamp_len=CdsShortTimestamp.TIMESTAMP_SIZE
            )
        except ValueError as e:
            _LOGGER.warning("Could not generate PUS TM object from raw data")
            _LOGGER.warning(f"Raw Packet: [{packet.hex(sep=',')}], REPR: {packet!r}")
            raise e
        service = tm_packet.service
        dedicated_handler = False
        if service == 1:
            verif_tm = Service1Tm.unpack(
                data=packet, params=UnpackParams(CdsShortTimestamp.TIMESTAMP_SIZE, 1, 2)
            )
            res = self.verif_wrapper.add_tm(verif_tm)
            if res is None:
                _LOGGER.info(
                    f"Received Verification TM[{tm_packet.service},"
                    f" {tm_packet.subservice}] with Request ID"
                    f" {verif_tm.tc_req_id.as_u32():#08x}"
                )
                _LOGGER.warning(
                    f"No matching telecommand found for {verif_tm.tc_req_id}"
                )
            else:
                self.verif_wrapper.log_to_console(verif_tm, res)
                self.verif_wrapper.log_to_file(verif_tm, res)
            dedicated_handler = True
        if service == 5:
            event_tm = Service5Tm.unpack(
                packet, timestamp_len=CdsShortTimestamp.TIMESTAMP_SIZE
            )
            _LOGGER.info(
                f"Received event packet TM [{event_tm.service}, {event_tm.subservice}]"
            )
        if service == 17:
            ping_tm = Service17Tm.unpack(
                packet, timestamp_len=CdsShortTimestamp.TIMESTAMP_SIZE
            )
            dedicated_handler = True
            if ping_tm.subservice == 2:
                _LOGGER.info("Received Ping Reply TM[17,2]")
            else:
                _LOGGER.info(
                    "Received Test Packet with unknown subservice"
                    f" {tm_packet.subservice}"
                )
        if tm_packet is None:
            _LOGGER.info(
                f"The service {service} is not implemented in Telemetry Factory"
            )
            tm_packet = PusTelemetry.unpack(
                packet, timestamp_len=CdsShortTimestamp.TIMESTAMP_SIZE
            )
        # TODO: Insert this into a DB instead. Maybe use sqlite for first variant.
        # self.raw_logger.log_tm(tm_packet)
        if not dedicated_handler:
            _LOGGER.info(
                f"Received PUS TM [{tm_packet.service}, {tm_packet.subservice}] with not dedicated "
                f"handler"
            )


class TcHandler(TcHandlerBase):
    def __init__(
        self,
        seq_count_provider: FileSeqCountProvider,
        verif_wrapper: VerificationWrapper,
    ):
        super(TcHandler, self).__init__()
        self.seq_count_provider = seq_count_provider
        self.verif_wrapper = verif_wrapper
        self.queue_helper = DefaultPusQueueHelper(
            queue_wrapper=QueueWrapper.empty(),
            seq_cnt_provider=seq_count_provider,
            pus_verificator=self.verif_wrapper.pus_verificator,
            tc_sched_timestamp_len=7,
            default_pus_apid=EXAMPLE_PUS_APID,
        )

    def send_cb(self, send_params: SendCbParams):
        entry_helper = send_params.entry
        if entry_helper.is_tc:
            if entry_helper.entry_type == TcQueueEntryType.PUS_TC:
                pus_tc_wrapper = entry_helper.to_pus_tc_entry()
                pus_tc_wrapper.pus_tc.seq_count = (
                    self.seq_count_provider.get_and_increment()
                )
                self.verif_wrapper.add_tc(pus_tc_wrapper.pus_tc)
                raw_tc = pus_tc_wrapper.pus_tc.pack()
                _LOGGER.info(f"Sending {pus_tc_wrapper.pus_tc}")
                send_params.com_if.send(raw_tc)
        elif entry_helper.entry_type == TcQueueEntryType.LOG:
            log_entry = entry_helper.to_log_entry()
            _LOGGER.info(log_entry.log_str)

    def queue_finished_cb(self, helper: ProcedureWrapper):
        if helper.proc_type == TcProcedureType.TREE_COMMANDING:
            def_proc = helper.to_tree_commanding_procedure()
            _LOGGER.info(f"Queue handling finished for command {def_proc.cmd_path}")

    def feed_cb(self, helper: ProcedureWrapper, wrapper: FeedWrapper):
        self.queue_helper.queue_wrapper = wrapper.queue_wrapper
        if helper.proc_type == TcProcedureType.TREE_COMMANDING:
            def_proc = helper.to_tree_commanding_procedure()
            cmd_path = def_proc.cmd_path
            assert cmd_path is not None
            # Path starts with / so the first entry of the list will be an empty string. We cut
            # off that string.
            cmd_path_list = cmd_path.split("/")[1:]
            if cmd_path_list[0] == "ping":
                return self.queue_helper.add_pus_tc(
                    PusTelecommand(apid=EXAMPLE_PUS_APID, service=17, subservice=1)
                )
            elif cmd_path_list[0] == "test":
                if cmd_path_list[1] == "event":
                    return self.queue_helper.add_pus_tc(
                        PusTelecommand(
                            apid=EXAMPLE_PUS_APID, service=17, subservice=128
                        )
                    )


# Note about lint disable: I could split up the function but I prefer to have the whole
# running main in one block for the example.
def main():  # noqa: C901
    # Apply the library logger formatting to the own root logger. Library logs will be propagated
    # to application logger.
    add_colorlog_console_logger(_LOGGER)
    tmtccmd.init_printout(False)
    hook_obj = ExampleHookClass(json_cfg_path=default_json_path())
    parser_wrapper = PreArgsParsingWrapper()
    parser_wrapper.create_default_parent_parser()
    parser_wrapper.create_default_parser()
    parser_wrapper.add_def_proc_args()
    params = SetupParams()
    post_args_wrapper = parser_wrapper.parse(hook_obj=hook_obj, setup_params=params)
    proc_wrapper = ProcedureParamsWrapper()
    if post_args_wrapper.use_gui:
        post_args_wrapper.set_params_without_prompts(proc_wrapper)
    else:
        post_args_wrapper.set_params_with_prompts(proc_wrapper)
    params.apid = EXAMPLE_PUS_APID
    if params.cmd_params.print_tree and not params.use_gui:
        perform_tree_printout(params.cmd_params, hook_obj.get_command_definitions())
        sys.exit(0)
    setup_args = SetupWrapper(
        hook_obj=hook_obj, setup_params=params, proc_param_wrapper=proc_wrapper
    )
    # Create console logger helper and file loggers
    tmtc_logger = RegularTmtcLogWrapper()
    printer = FsfwTmTcPrinter(tmtc_logger.logger)
    verificator = PusVerificator()
    verification_wrapper = VerificationWrapper(
        verificator, _LOGGER, printer.file_logger
    )
    # Create primary TM handler and add it to the CCSDS Packet Handler
    tm_handler = PusTmHandler(verification_wrapper, printer)
    ccsds_handler = CcsdsTmHandler(generic_handler=None)
    ccsds_handler.add_apid_handler(tm_handler)

    # Create TC handler
    seq_count_provider = PusFileSeqCountProvider()
    tc_handler = TcHandler(seq_count_provider, verification_wrapper)
    tmtccmd.setup(setup_args=setup_args)
    init_proc = params_to_procedure_conversion(setup_args.proc_param_wrapper)
    tmtc_backend = tmtccmd.create_default_tmtc_backend(
        setup_wrapper=setup_args,
        tm_handler=ccsds_handler,
        tc_handler=tc_handler,
        init_procedure=init_proc,
    )
    tmtccmd.start(tmtc_backend=tmtc_backend, hook_obj=hook_obj)
    try:
        while True:
            state = tmtc_backend.periodic_op(None)
            if state.request == BackendRequest.TERMINATION_NO_ERROR:
                tmtc_backend.close_com_if()
                sys.exit(0)
            elif state.request == BackendRequest.DELAY_IDLE:
                _LOGGER.info("TMTC Client in IDLE mode")
                time.sleep(3.0)
            elif state.request == BackendRequest.DELAY_LISTENER:
                time.sleep(0.8)
            elif state.request == BackendRequest.DELAY_CUSTOM:
                if state.next_delay.total_seconds() <= 0.4:
                    time.sleep(state.next_delay.total_seconds())
                else:
                    time.sleep(0.4)
            elif state.request == BackendRequest.CALL_NEXT:
                pass
    except KeyboardInterrupt:
        tmtc_backend.close_com_if()
        sys.exit(0)


if __name__ == "__main__":
    main()



from __future__ import annotations

import struct
from typing import Deque


from spacepackets.ecss.pus_1_verification import (
    Service1Tm,
)


class Service1FsfwWrapper:
    def __init__(self, tm: Service1Tm):
        self.tm = tm
        if tm.has_failure_notice and tm.failure_notice is not None:
            self.error_param_1 = struct.unpack("!I", tm.failure_notice.data[0:4])[0]
            self.error_param_2 = struct.unpack("!I", tm.failure_notice.data[4:8])[0]


PusVerifQueue = Deque[Service1Tm]



"""Hamming Code Implementation
Hamming codes belong to the family of linear error correcting codes.
Documentation: https://en.wikipedia.org/wiki/Hamming_code
They can be used to identify up to two bit errors and correct one bit error per 256 byte block.
"""

import logging
from enum import Enum

# Translated from ATMEL C library.
# /* ----------------------------------------------------------------------------
# *         ATMEL Microcontroller Software Support
# * ----------------------------------------------------------------------------
# * Copyright (c) 2008, Atmel Corporation
# *
# * All rights reserved.
# *
# * Redistribution and use in source and binary forms, with or without
# * modification, are permitted provided that the following conditions are met:
# *
# * - Redistributions of source code must retain the above copyright notice,
# * this list of conditions and the disclaimer below.
# *
# * Atmel's name may not be used to endorse or promote products derived from
# * this software without specific prior written permission.
# *
# * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR
# * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
# * DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT, INDIRECT,
# * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# * ----------------------------------------------------------------------------
# */


_LOGGER = logging.getLogger(__name__)


class HammingReturnCodes(Enum):
    # No bit flips
    CODE_OKAY = 0
    # Single bit flip which can be corrected
    ERROR_SINGLE_BIT = 1
    # Error in the hamming code
    ERROR_ECC = 2
    # Multi bit error which can not be corrected
    ERROR_MULTI_BIT = 3
    # Invalid input
    OTHER_ERROR = 4


def hamming_compute_256x(data: bytearray) -> bytearray:
    """Computes 3-bytes hamming codes for a data block whose size is multiple of
    256 bytes. Each 256 bytes block gets its own code.

    :param data: Data to compute code for. Should be a multiple of 256 bytes, pad data with 0
        if necessary!
    :return: bytearray of hamming codes with the size (3 / 256 * size). Empty bytearray if input
        is invalid.
    """
    if len(data) % 256 != 0:
        _LOGGER.error(
            "hamming_compute_256: Invalid input, datablock is not a multiple of "
            "256 bytes!"
        )
        return bytearray()

    remaining_size = len(data)
    hamming_code = bytearray()
    current_idx = 0
    while remaining_size > 0:
        hamming_code.extend(hamming_compute_256(data[current_idx : current_idx + 256]))
        remaining_size -= 256
        current_idx += 256
    return hamming_code


def hamming_verify_256x(
    data: bytearray, original_hamming_code: bytearray
) -> HammingReturnCodes:
    if len(data) % 256 != 0:
        _LOGGER.error(
            "hamming_compute_256: Invalid input, datablock is not a multiple of "
            "256 bytes!"
        )
        return HammingReturnCodes.OTHER_ERROR
    if len(original_hamming_code) != len(data) / 256 * 3:
        _LOGGER.error(
            "hamming_compute_256: Invalid input, original hamming code does not have"
            " thecorrect size!"
        )
        return HammingReturnCodes.OTHER_ERROR

    remaining_size = len(data)
    current_data_idx = 0
    current_hamming_idx = 0
    error_code = HammingReturnCodes.CODE_OKAY
    while remaining_size > 0:
        current_data = bytearray(data[current_data_idx : current_data_idx + 256])
        error_code = hamming_verify_256(
            current_data,
            original_hamming_code[current_hamming_idx : current_hamming_idx + 3],
        )
        if error_code == HammingReturnCodes.ERROR_SINGLE_BIT:
            # Assign corrected data
            data[current_data_idx : current_data_idx + 256] = current_data
            _LOGGER.info(
                "Corrected single bit error at data block starting at"
                f" {current_data_idx}"
            )
            error_code = HammingReturnCodes.ERROR_SINGLE_BIT
        elif error_code == HammingReturnCodes.ERROR_MULTI_BIT:
            _LOGGER.info(
                f"Detected multi-bit error at data block starting at {current_data_idx}"
            )
            return error_code
        elif error_code == HammingReturnCodes.ERROR_ECC:
            _LOGGER.info("Possible error in ECC code")
            return error_code
        current_data_idx += 256
        current_hamming_idx += 3
        remaining_size -= 256
    return error_code


def hamming_compute_256(data: bytearray) -> bytearray:  # noqa: C901
    """Takes a bytearray with the size of 256 bytes and calculates the 22 parity bits for the
    hamming code which will be returned as a three byte bytearray.

    :param data:
    :return:
    """
    hamming_code = bytearray(3)
    if len(data) != 256:
        _LOGGER.error(
            "hamming_compute_256: Invalid input, data does not have "
            "a length of 256 bytes!"
        )
        return hamming_code

    # Xor all bytes together to get the column sum;
    # At the same time, calculate the even and odd line codes
    column_sum = 0
    even_line_code = 0
    odd_line_code = 0
    even_column_code = 0
    odd_column_code = 0
    for index, byte in enumerate(data):
        column_sum ^= byte
        if (bin(byte).count("1") & 1) == 1:
            """
            Parity groups are formed by forcing a particular index bit to 0
            (even) or 1 (odd).
            Example on one byte:

            bits (dec)  7   6   5   4   3   2   1   0
                 (bin) 111 110 101 100 011 010 001 000
                                       '---'---'---'----------.
                                                              |
            groups P4' ooooooooooooooo eeeeeeeeeeeeeee P4     |
                   P2' ooooooo eeeeeee ooooooo eeeeeee P2     |
                   P1' ooo eee ooo eee ooo eee ooo eee P1     |
                                                              |
            We can see that:                                  |
             - P4  -> bit 2 of index is 0 --------------------'
             - P4' -> bit 2 of index is 1.
             - P2  -> bit 1 of index if 0.
             - etc...
            We deduce that a bit position has an impact on all even Px if
            the log2(x)nth bit of its index is 0
                ex: log2(4) = 2, bit2 of the index must be 0 (-> 0 1 2 3)
            and on all odd Px' if the log2(x)nth bit of its index is 1
                ex: log2(2) = 1, bit1 of the index must be 1 (-> 0 1 4 5)

            As such, we calculate all the possible Px and Px' values at the
            same time in two variables, evenLineCode and oddLineCode, such as
                 evenLineCode bits: P128  P64  P32  P16  P8  P4  P2  P1
                 oddLineCode  bits: P128' P64' P32' P16' P8' P4' P2' P1'
            """
            even_line_code ^= 255 - index
            odd_line_code ^= index

    # At this point, we have the line parities, and the column sum.
    # First, We must caculate the parity group values on the column sum.
    for index in range(8):
        if column_sum & 1:
            even_column_code ^= 7 - index
            odd_column_code ^= index
        column_sum >>= 1

    # Now, we must interleave the parity values, to obtain the following layout:
    # Code[0] = Line1
    # Code[1] = Line2
    # Code[2] = Column
    # Line = Px' Px P(x-1)- P(x-1) ...
    # Column = P4' P4 P2' P2 P1' P1 PadBit PadBit
    hamming_code[0] = 0
    hamming_code[1] = 0
    hamming_code[2] = 0

    for index in range(4):
        hamming_code[0] <<= 2
        hamming_code[1] <<= 2
        hamming_code[2] <<= 2

        # Line 1
        if (odd_line_code & 0x80) != 0:
            hamming_code[0] |= 2
        if (even_line_code & 0x80) != 0:
            hamming_code[0] |= 1

        # Line 2
        if (odd_line_code & 0x08) != 0:
            hamming_code[1] |= 2
        if (even_line_code & 0x08) != 0:
            hamming_code[1] |= 1

        # Column
        if (odd_column_code & 0x04) != 0:
            hamming_code[2] |= 2
        if (even_column_code & 0x04) != 0:
            hamming_code[2] |= 1

        odd_line_code <<= 1
        even_line_code <<= 1
        odd_column_code <<= 1
        even_column_code <<= 1

    # Invert codes (linux compatibility)
    hamming_code[0] ^= 0xFF
    hamming_code[1] ^= 0xFF
    hamming_code[2] ^= 0xFF

    return hamming_code


def hamming_verify_256(
    data: bytearray, original_hamming_code: bytearray
) -> HammingReturnCodes:
    """Verifies and corrects a 256-bytes block of data using the given 22-bits hamming code.
    Returns 0 if there is no error, otherwise returns a HAMMING_ERROR code.

    :param data: 256 code block to verify
    :param original_hamming_code: Original 3 byte hamming code with 22 parity bits
    :return: See HammingReturnCodes enums.
     - -1 for invalid input
     - 0 if there are no errors.
     - 1 if there is a single bit error which has been corrected
     - 2 if the hamming code has been corrupted
     - 3 if there was a multi bit error which can not be corrected
    """
    if len(data) != 256:
        _LOGGER.error(
            "hamming_compute_256: Invalid input, data does not have "
            "a length of 256 bytes!"
        )
        return HammingReturnCodes.OTHER_ERROR
    if len(original_hamming_code) != 3:
        _LOGGER.error(
            "hamming_compute_256: Invalid input, hamming code does not have "
            "a length of 3 bytes!"
        )
        return HammingReturnCodes.OTHER_ERROR

    # Calculate new code
    computed_hamming_code = hamming_compute_256(data)
    correction_code = bytearray(3)

    # Xor both codes together
    correction_code[0] = original_hamming_code[0] ^ computed_hamming_code[0]
    correction_code[1] = original_hamming_code[1] ^ computed_hamming_code[1]
    correction_code[2] = original_hamming_code[2] ^ computed_hamming_code[2]

    # If all bytes are 0, there is not error
    if correction_code[0] == 0 and correction_code[1] == 0 and correction_code[2] == 0:
        return HammingReturnCodes.CODE_OKAY

    # If there is a single bit error, there are 11 bits set to 1
    hamming_bit_count = (
        bin(correction_code[0]).count("1")
        + bin(correction_code[1]).count("1")
        + bin(correction_code[2]).count("1")
    )
    if hamming_bit_count == 11:
        # Get byte and bit indexes
        byte_idx = correction_code[0] & 0x80
        byte_idx |= (correction_code[0] << 1) & 0x40
        byte_idx |= (correction_code[0] << 2) & 0x20
        byte_idx |= (correction_code[0] << 3) & 0x10

        byte_idx |= (correction_code[1] >> 4) & 0x08
        byte_idx |= (correction_code[1] >> 3) & 0x04
        byte_idx |= (correction_code[1] >> 2) & 0x02
        byte_idx |= (correction_code[1] >> 1) & 0x01

        bit_idx = (correction_code[2] >> 5) & 0x04
        bit_idx |= (correction_code[2] >> 4) & 0x02
        bit_idx |= (correction_code[2] >> 3) & 0x01

        # Correct bit
        print_string = "Correcting byte " + str(byte_idx) + " at bit " + str(bit_idx)
        _LOGGER.info(print_string)
        data[byte_idx] ^= 1 << bit_idx
        return HammingReturnCodes.ERROR_SINGLE_BIT

    # Check whether ECC has been corrupted
    if hamming_bit_count == 1:
        return HammingReturnCodes.ERROR_ECC
    # Otherwise, this is a multibit error
    else:
        return HammingReturnCodes.ERROR_MULTI_BIT


def hamming_test():
    """Algorithm was verified with this  simple test."""
    test_data = bytearray(256)
    for index in range(128):
        test_data[index] = index
    for index in range(128):
        test_data[index + 128] = 128 - index
    hamming_code = hamming_compute_256(test_data)
    print(
        "Hamming code: "
        + str(hex(hamming_code[0]))
        + ", "
        + str(hex(hamming_code[1]))
        + ", "
        + str(hex(hamming_code[2]))
    )



#:wq/usr/bin/env python3
from unittest import TestCase

from spacepackets.ccsds.time import CdsShortTimestamp
from spacepackets.ecss.pus_17_test import Service17Tm


class TestTelemetry(TestCase):
    def setUp(self) -> None:
        self.apid = 0xEF
        self.pus_17_telemetry = Service17Tm(
            apid=self.apid,
            subservice=1,
            ssc=36,
            timestamp=CdsShortTimestamp.now().pack(),
        )

        self.pus_17_raw = self.pus_17_telemetry.pack()

    def test_generic_pus_c(self):
        def tm_func(raw_telemetry: bytearray):
            return Service17Tm.unpack(data=raw_telemetry, timestamp_len=7)

        self.assertRaises(ValueError, tm_func, bytearray())
        self.assertRaises(ValueError, tm_func, None)

        pus_17_telemetry = Service17Tm.unpack(data=self.pus_17_raw, timestamp_len=7)
        self.assertTrue(pus_17_telemetry.service == 17)
        self.assertTrue(pus_17_telemetry.apid == 0xEF)
        self.assertTrue(pus_17_telemetry.subservice == 1)
        self.assertTrue(pus_17_telemetry.seq_count == 36)
        self.assertTrue(pus_17_telemetry.source_data == bytearray())
        self.assertTrue(pus_17_telemetry.pus_tm.packet_id.raw() == 0x8 << 8 | 0xEF)



from importlib import metadata


def get_version() -> str:
    """Retrieve the package version using the
    `importlib.metadata API <https://docs.python.org/3/library/importlib.metadata.html>`_.
    """
    return metadata.version("tmtccmd")



import struct

from deprecated.sphinx import deprecated
from spacepackets.ecss import PusTelecommand

from tmtccmd.pus.s8_fsfw_action_defs import CustomSubservice


@deprecated(
    version="v4.0.0a2",
    reason="use create... API instead",
)
def make_fsfw_action_cmd(
    object_id: bytes,
    action_id: int,
    user_data: bytes = bytes(),
    apid: int = 0,
    seq_count: int = 0,
) -> PusTelecommand:
    return create_action_cmd(
        object_id=object_id,
        action_id=action_id,
        user_data=user_data,
        apid=apid,
        seq_count=seq_count,
    )


def create_action_cmd(
    object_id: bytes,
    action_id: int,
    user_data: bytes = bytes(),
    apid: int = 0,
    seq_count: int = 0,
) -> PusTelecommand:
    data_to_pack = bytearray(object_id)
    data_to_pack += make_action_id(action_id) + user_data
    return PusTelecommand(
        service=8,
        subservice=CustomSubservice.TC_FUNCTIONAL_CMD,
        app_data=data_to_pack,
        apid=apid,
        seq_count=seq_count,
    )


def make_action_id(action_id: int) -> bytearray:
    return bytearray(struct.pack("!I", action_id))



import collections.abc
import logging
from typing import Tuple, Union
from contextlib import contextmanager


def wrapped_prompt(text: str):
    return input(text)


class AnsiColors:
    RED = "\x1b[31m"
    GREEN = "\x1b[32m"
    YELLOW = "\x1b[33m"
    BLUE = "\x1b[34m"
    MAGNETA = "\x1b[35m"
    CYAN = "\x1b[36m"
    RESET = "\x1b[0m"
    BOLD = "\033[1m"


def check_args_in_dict(
    param: any, iterable: Union[collections.abc.Iterable, dict], warning_hint: str
) -> Tuple[bool, int]:
    """This functions checks whether the integer representation of a given parameter in
    contained within the passed collections, for example an (integer) enumeration.
    Please note that if the passed parameter has a string representation but is a digit,
    this function will attempt to check whether the integer representation is contained
    inside the passed enumeration.
    :param param:           Value to be checked
    :param iterable:     Enumeration, for example a enum.Enum or enum.IntEnum implementation
    :param warning_hint:
    :return:
    """
    might_be_integer = False
    if param is not None:
        if isinstance(param, str):
            if param.isdigit():
                might_be_integer = True
        elif isinstance(param, int):
            pass
        else:
            logging.getLogger(__name__).warning(f"Passed {warning_hint} type invalid.")
            return False, 0
    else:
        logging.getLogger(__name__).warning(f"No {warning_hint} argument passed.")
        return False, 0

    res_tuple = False, 0
    if isinstance(iterable, dict):
        for idx, enum_value in iterable.items():
            if param == enum_value:
                return True, idx
    else:
        res_tuple = __handle_iterable_non_dict(
            param=param,
            iterable=iterable,
            might_be_integer=might_be_integer,
            init_res_tuple=res_tuple,
        )
    return res_tuple


def __handle_iterable_non_dict(
    param: any,
    iterable: collections.abc.Iterable,
    might_be_integer: bool,
    init_res_tuple: Tuple[bool, any],
) -> (bool, any):
    param_list = list()
    for idx, enum_value in enumerate(iterable):
        if isinstance(enum_value.value, str):
            # Make this case insensitive
            param_list.append(enum_value.value.lower())
        else:
            param_list.append(enum_value.value)
    if param not in param_list:
        if might_be_integer:
            if int(param) in param_list:
                return True, int(param)
        return False, 0
    return init_res_tuple


@contextmanager
def acquire_timeout(lock, timeout):
    """Helper functions which allows to check result of the acquire operation while also
    using the context manager.
    :param lock:
    :param timeout:
    :return:
    """
    result = lock.acquire(timeout=timeout)
    yield result
    if result:
        lock.release()



import enum
import pprint

from deprecated.sphinx import deprecated
from tmtccmd.core.globals_manager import update_global
from tmtccmd.config.defs import (
    CoreModeList,
    CoreComInterfaces,
)
from tmtccmd.config.tmtc import TmtcDefinitionWrapper

DEF_WRAPPER = None


class CoreGlobalIds(enum.IntEnum):
    """
    Numbers from 128 to 200 are reserved for core globals
    """

    # Object handles
    TMTC_HOOK = 128
    COM_INTERFACE_HANDLE = 129
    TM_LISTENER_HANDLE = 130
    TMTC_PRINTER_HANDLE = 131
    TM_HANDLER_HANDLE = 132
    PRETTY_PRINTER = 133

    # Parameters
    JSON_CFG_PATH = 139
    MODE = 141
    CURRENT_SERVICE = 142
    COM_IF = 144
    OP_CODE = 145
    TM_TIMEOUT = 146
    SERVICE_OP_CODE_DICT = 147
    COM_IF_DICT = 148

    # Miscellaneous
    DISPLAY_MODE = 150
    USE_LISTENER_AFTER_OP = 151
    PRINT_HK = 152
    PRINT_TM = 153
    PRINT_RAW_TM = 154
    PRINT_TO_FILE = 155
    RESEND_TC = 156
    TC_SEND_TIMEOUT_FACTOR = 157

    # Config dictionaries
    USE_SERIAL = 160
    SERIAL_CONFIG = 161
    USE_ETHERNET = 162
    ETHERNET_CONFIG = 163
    END = 300


@deprecated(version="6.0.0rc0", reason="globals module deprecated")
def set_default_globals_pre_args_parsing(
    apid: int,
    com_if_id: str = CoreComInterfaces.DUMMY.value,
    custom_com_if_dict=None,
    display_mode="long",
    tm_timeout: float = 4.0,
    print_to_file: bool = True,
    tc_send_timeout_factor: float = 2.0,
):
    if custom_com_if_dict is None:
        custom_com_if_dict = dict()
    update_global(CoreGlobalIds.COM_IF, com_if_id)
    update_global(CoreGlobalIds.TC_SEND_TIMEOUT_FACTOR, tc_send_timeout_factor)
    update_global(CoreGlobalIds.TM_TIMEOUT, tm_timeout)
    update_global(CoreGlobalIds.DISPLAY_MODE, display_mode)
    update_global(CoreGlobalIds.PRINT_TO_FILE, print_to_file)
    update_global(CoreGlobalIds.CURRENT_SERVICE, 17)
    update_global(CoreGlobalIds.SERIAL_CONFIG, dict())
    update_global(CoreGlobalIds.ETHERNET_CONFIG, dict())
    pp = pprint.PrettyPrinter()
    update_global(CoreGlobalIds.PRETTY_PRINTER, pp)
    update_global(CoreGlobalIds.TM_LISTENER_HANDLE, None)
    update_global(CoreGlobalIds.COM_INTERFACE_HANDLE, None)
    update_global(CoreGlobalIds.TMTC_PRINTER_HANDLE, None)
    update_global(CoreGlobalIds.PRINT_RAW_TM, False)
    update_global(CoreGlobalIds.USE_LISTENER_AFTER_OP, True)
    update_global(CoreGlobalIds.RESEND_TC, False)
    update_global(CoreGlobalIds.OP_CODE, "0")
    update_global(CoreGlobalIds.MODE, CoreModeList.LISTENER_MODE)


@deprecated(version="6.0.0rc0", reason="globals module deprecated")
def check_and_set_other_args(args):
    if args.listener is not None:
        update_global(CoreGlobalIds.USE_LISTENER_AFTER_OP, args.listener)
    if args.tm_timeout is not None:
        update_global(CoreGlobalIds.TM_TIMEOUT, args.tm_timeout)
    if args.print_hk is not None:
        update_global(CoreGlobalIds.PRINT_HK, args.print_hk)
    if args.print_tm is not None:
        update_global(CoreGlobalIds.PRINT_TM, args.print_tm)
    if args.raw_print is not None:
        update_global(CoreGlobalIds.PRINT_RAW_TM, args.raw_print)
    if args.print_log is not None:
        update_global(CoreGlobalIds.PRINT_TO_FILE, args.print_log)
    if args.resend_tc is not None:
        update_global(CoreGlobalIds.RESEND_TC, args.resend_tc)
    update_global(CoreGlobalIds.TC_SEND_TIMEOUT_FACTOR, 3)


@deprecated(version="8.0.0", reason="use command tree API instead")
def get_default_tmtc_defs() -> TmtcDefinitionWrapper:
    global DEF_WRAPPER
    if DEF_WRAPPER is None:
        DEF_WRAPPER = TmtcDefinitionWrapper()
    return DEF_WRAPPER



import os
from pathlib import Path
from unittest import TestCase

from spacepackets.ccsds.time import CdsShortTimestamp
from spacepackets.ecss.pus_1_verification import (
    RequestId,
    VerificationParams,
    Subservice,
    Service1Tm,
)

from tmtccmd.pus.s17_test import create_service_17_ping_command
from tmtccmd.logging import LOG_DIR
from tmtccmd.logging.pus import (
    RegularTmtcLogWrapper,
    RawTmtcRotatingLogWrapper,
)


# TODO: Use temp files to test loggers?
class TestPrintersLoggers(TestCase):
    def setUp(self):
        self.apid = 0x07
        self.log_path = Path(LOG_DIR)
        if not self.log_path.exists():
            self.log_path.mkdir()
        self.regular_file_name = Path(
            RegularTmtcLogWrapper.get_current_tmtc_file_name()
        )

    def test_pus_loggers(self):
        regular_tmtc_logger = RegularTmtcLogWrapper(self.regular_file_name)
        raw_tmtc_log = RawTmtcRotatingLogWrapper(max_bytes=1024, backup_count=10)
        pus_tc = create_service_17_ping_command()
        raw_tmtc_log.log_tc(pus_tc)
        pus_tm = Service1Tm(
            apid=self.apid,
            subservice=Subservice.TM_START_SUCCESS,
            timestamp=CdsShortTimestamp.now().pack(),
            verif_params=VerificationParams(
                req_id=RequestId(pus_tc.packet_id, pus_tc.packet_seq_control)
            ),
        )
        raw_tmtc_log.log_tm(pus_tm.pus_tm)
        self.assertTrue(Path(self.regular_file_name).exists())
        regular_tmtc_logger.logger.info("Test")
        # There should be 2 files now because 1024 bytes are not enough to accomate all info
        self.assertTrue(Path(raw_tmtc_log.file_name).exists())
        self.assertTrue(Path(f"{raw_tmtc_log.file_name}.1").exists())

    def test_print_functions(self):
        pass

    def tearDown(self):
        """Reset the hook object"""
        if self.regular_file_name.exists():
            os.remove(self.regular_file_name)



from typing import List
from collections import deque
from PyQt6.QtCore import QModelIndex, pyqtSignal
from PyQt6.QtGui import QStandardItem, QStandardItemModel
from PyQt6.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QLabel,
    QListWidget,
    QListWidgetItem,
    QPushButton,
    QTreeView,
    QVBoxLayout,
    QWidget,
)

from tmtccmd.config.tmtc import CmdTreeNode


class CommandPathSelectWidget(QWidget):
    NODE_NAME_COLUMN = 0
    DESCRIPTION_COLUMN = 1

    path_selected_sig = pyqtSignal(str)
    closed = pyqtSignal()

    def __init__(self, root_node: CmdTreeNode, last_selected_items: deque):
        super().__init__()
        self.num_of_display_last_sel_items = 10
        self.last_selected_items = last_selected_items
        self.root_node = root_node
        self.currently_selected_path = None
        self.init_ui()

    def init_ui(self):
        self.setWindowTitle("Command path selector")

        self.tree_model = QStandardItemModel()
        self.tree_model.setHorizontalHeaderLabels(["Node Name", "Description"])
        parent_item = self.tree_model.invisibleRootItem()
        assert parent_item is not None
        CommandPathSelectWidget.build_tree_model_recursively(
            parent_item, self.root_node
        )

        self.tree_view = QTreeView()
        self.tree_view.setModel(self.tree_model)
        self.tree_view.expanded.connect(self._on_item_expanded)
        self.tree_view.doubleClicked.connect(self._on_tree_view_double_click)

        self.expand_full_button = QPushButton("Expand All")
        self.expand_full_button.clicked.connect(self.expand_all_items)

        self.fold_all_button = QPushButton("Fold All")
        self.fold_all_button.clicked.connect(self.collapse_all_items)

        button_layout = QHBoxLayout()
        button_layout.addWidget(self.expand_full_button)
        button_layout.addWidget(self.fold_all_button)

        self.last_sel_items_button = QPushButton("Last Selected Items")
        self.last_sel_items_button.clicked.connect(self._last_sel_items_clicked)
        self.select_path_button = QPushButton("Select Current Path")
        self.select_path_button.clicked.connect(self._select_path_clicked)
        self.copy_path_button = QPushButton("Copy Current Path to Clipboard")
        self.copy_path_button.clicked.connect(self._copy_path_clicked)
        self.path_confirmed_button = QPushButton("Confirm Selected Path")
        self.path_confirmed_button.clicked.connect(self._path_confirmed_clicked)

        self.path_label = QLabel()

        layout = QVBoxLayout()
        layout.addLayout(button_layout)
        layout.addWidget(self.tree_view)

        layout.addWidget(self.last_sel_items_button)
        layout.addWidget(self.select_path_button)
        layout.addWidget(self.copy_path_button)
        layout.addWidget(self.path_label)
        layout.addWidget(self.path_confirmed_button)
        self.setLayout(layout)

    def expand_all_items(self):
        self.tree_view.expandAll()

    def collapse_all_items(self):
        self.tree_view.collapseAll()

    def _last_sel_items_clicked(self):
        self.last_sel_cmds_widget = QListWidget()
        self.last_sel_cmds_widget.setWindowTitle("Last selected items")
        for item in self.last_selected_items:
            self.last_sel_cmds_widget.addItem(QListWidgetItem(item))
        self.last_sel_cmds_widget.doubleClicked.connect(self._last_sel_item_confirmed)
        self.last_sel_cmds_widget.show()

    def _path_confirmed_clicked(self):
        if self.currently_selected_path is None:
            # TODO: Error handling in some shape or form, maybe message box?
            return
        self.path_selected_sig.emit(self.currently_selected_path)
        if len(self.last_selected_items) >= self.num_of_display_last_sel_items:
            self.last_selected_items.popleft()
        self.last_selected_items.append(self.currently_selected_path)
        self.closed.emit()
        self.close()

    @staticmethod
    def build_tree_model_recursively(
        current_item: QStandardItem, current_node: CmdTreeNode
    ):
        for child in current_node.children.values():
            new_item = QStandardItem(child.name)
            current_item.appendRow([new_item, QStandardItem(child.description)])
            if child.children:
                CommandPathSelectWidget.build_tree_model_recursively(new_item, child)

    def _on_tree_view_double_click(self, _index: int):
        selected_indexes = self.tree_view.selectedIndexes()
        if selected_indexes:
            index = selected_indexes[0]  # Get the first selected index
            item = self.tree_model.itemFromIndex(index)
            assert item is not None
            if not item.hasChildren():
                self._select_path_clicked_by_index(selected_indexes)
            else:
                self.tree_view.expand(selected_indexes[0])
        self._path_confirmed_clicked()

    def _on_item_expanded(self, _index: int):
        self.tree_view.resizeColumnToContents(CommandPathSelectWidget.NODE_NAME_COLUMN)
        self.tree_view.resizeColumnToContents(
            CommandPathSelectWidget.DESCRIPTION_COLUMN
        )
        self.resize(
            self.tree_view.sizeHintForColumn(CommandPathSelectWidget.NODE_NAME_COLUMN)
            + self.tree_view.sizeHintForColumn(
                CommandPathSelectWidget.DESCRIPTION_COLUMN
            )
            + 40,
            self.size().height(),
        )

    def _select_path_clicked(self):
        self._select_path_clicked_by_index(self.tree_view.selectedIndexes())

    def _select_path_clicked_by_index(self, selected_indexes: List[QModelIndex]):
        if selected_indexes:
            index = selected_indexes[0]  # Get the first selected index
            item = self.tree_model.itemFromIndex(index)
            self._update_sel_path_common(f"/{self._get_item_path(item)}")

    def _last_sel_item_confirmed(self, item: QModelIndex):
        self._update_sel_path_common(self.last_selected_items[item.row()])
        self.last_sel_cmds_widget.close()

    def _update_sel_path_common(self, path: str):
        self.currently_selected_path = path
        self.path_label.setText(f"Selected: {self.currently_selected_path}")

    def _copy_path_clicked(self):
        if self.currently_selected_path is None:
            self._select_path_clicked()
        if self.currently_selected_path is not None:
            print(f"Copied path {self.currently_selected_path} to clipboard.")
            clipboard = QApplication.clipboard()
            if clipboard is not None:
                clipboard.setText(self.currently_selected_path)

    def _get_item_path(self, item):
        path = []
        while item:
            path.insert(0, item.text())
            item = item.parent()
        return "/".join(path)



"""Contains the TmListener which can be used to listen to Telemetry in the background"""

from typing import Dict, List, Tuple

from spacepackets.ccsds.spacepacket import get_apid_from_raw_space_packet

from tmtccmd.tmtc.common import TelemetryQueueT, CcsdsTmHandler
from tmtccmd.com import ComInterface


INVALID_APID = -2
UNKNOWN_TARGET_ID = -1
QueueDictT = Dict[int, Tuple[TelemetryQueueT, int]]
QueueListT = List[Tuple[int, TelemetryQueueT]]


class PacketsTooSmallForCcsds(Exception):
    def __init__(self, packets: List[bytes]):
        self.packets = packets


class CcsdsTmListener:
    """Simple helper object which can be used for retrieving and routing CCSDS packets.
    It can be used to poll CCSDS packets from a provided :py:class:`tmtccmd.com_if.ComInterface`
    and then route them using a provided CCSDS TM handler.
    """

    def __init__(
        self,
        tm_handler: CcsdsTmHandler,
    ):
        """Initiate a TM listener.

        :param tm_handler: If valid CCSDS packets are found, they are dispatched to
            the passed handler
        """
        self.__tm_handler = tm_handler

    def operation(self, com_if: ComInterface) -> int:
        """Core operation to route packet to the provided handler.

        :param com_if:
        :raises PacketsTooSmallForCcsds: If any of the received packets are too small.
            The internal handler will still continue to process the remaining packet list retrieved
            from the COM interface.
        :return:
        """
        packet_list = com_if.receive()
        for tm_packet in packet_list:
            self.__handle_ccsds_space_packet(tm_packet)
        return len(packet_list)

    def __handle_ccsds_space_packet(self, tm_packet: bytes):
        invalid_packets = []
        if len(tm_packet) < 6:
            invalid_packets.append(tm_packet)
        else:
            apid = get_apid_from_raw_space_packet(tm_packet)
            self.__tm_handler.handle_packet(apid, tm_packet)
            return True
        if len(invalid_packets) > 0:
            raise PacketsTooSmallForCcsds(invalid_packets)



from abc import abstractmethod
from typing import Any, Optional

from .backend_state import BackendState


class BackendBase:
    @abstractmethod
    def open_com_if(self):
        """Start the backend. Raise RuntimeError on failure"""
        pass

    @abstractmethod
    def close_com_if(self):
        pass

    @abstractmethod
    def periodic_op(self, args: Optional[Any]) -> BackendState:
        pass



import dataclasses
import enum
from typing import Optional, Dict


class Severity(enum.IntEnum):
    INFO = 1
    LOW = (2,)
    MEDIUM = (3,)
    HIGH = 4


def str_to_severity(string: str) -> Optional[Severity]:
    if string == "INFO":
        return Severity.INFO
    elif string == "LOW":
        return Severity.LOW
    elif string == "MEDIUM":
        return Severity.MEDIUM
    elif string == "HIGH":
        return Severity.HIGH


@dataclasses.dataclass
class EventInfo:
    id: int = 0
    name: str = ""
    severity: str = ""
    info: str = ""
    file_location: str = ""


EventDictT = Dict[int, EventInfo]



import struct
from unittest import TestCase

from spacepackets.ecss import RequestId, PusTc, PacketFieldU8
from spacepackets.ecss.pus_1_verification import (
    Service1Tm,
    Subservice,
    VerificationParams,
    FailureNotice,
)
from tmtccmd.pus.tm.s1_verification import Service1FsfwWrapper


class TestVerif1TmWrapper(TestCase):
    def setUp(self) -> None:
        self.apid = 0x01
        self.ping_tc = PusTc(apid=self.apid, service=17, subservice=1)
        pass

    def test_basic(self):
        context_param_1 = bytearray([0x10, 0x20, 0x30, 0x40])
        context_param_1_as_int = struct.unpack("!I", context_param_1)[0]
        context_param_2 = bytearray([0x40, 0x30, 0x20, 0x10])
        context_param_2_as_int = struct.unpack("!I", context_param_2)[0]
        failure_notice = FailureNotice(
            code=PacketFieldU8(1), data=context_param_1 + context_param_2
        )
        service_1_tm = Service1Tm(
            apid=self.apid,
            subservice=Subservice.TM_START_FAILURE,
            timestamp=bytes(),
            verif_params=VerificationParams(
                step_id=None,
                req_id=RequestId.from_pus_tc(self.ping_tc),
                failure_notice=failure_notice,
            ),
        )
        wrapper = Service1FsfwWrapper(service_1_tm)
        self.assertEqual(wrapper.error_param_1, context_param_1_as_int)
        self.assertEqual(wrapper.error_param_2, context_param_2_as_int)



from __future__ import annotations
import enum
import struct

from spacepackets.ecss import PusTelecommand


class TypeOfTimeWindow(enum.IntEnum):
    SELECT_ALL = 0
    FROM_TIMETAG_TO_TIMETAG = 1
    FROM_TIMETAG = 2
    TO_TIMETAG = 3


class Subservice(enum.IntEnum):
    """Unless specified, TCs and TMs are related to a request ID"""

    TC_ENABLE = 1
    TC_DISABLE = 2
    TC_RESET = 3
    TC_INSERT = 4
    TC_DELETE = 5
    TC_DELETE_WITH_FILTER = 6
    TC_TIMESHIFT = 7
    TC_TIMESHIFT_WITH_FILTER = 8
    TC_DETAIL_REPORT_TIME_BASED = 9
    TM_DETAIL_REPORT_TIME_BASED = 10
    TC_DETAIL_REPORT_FILTER_BASED = 11
    TM_DETAIL_REPORT_FILTER_BASED = 12
    TC_TIMESHIFT_ALL = 15


class TcSchedReqId:
    def __init__(self, apid: int, seq_cnt: int, src_id: int):
        """
        :raises ValueError: Input invalid
        """
        if apid > pow(2, 11) or apid < 0:
            raise ValueError
        self.apid = apid
        if seq_cnt > pow(2, 16) or seq_cnt < 0:
            raise ValueError
        self.seq_cnt = seq_cnt
        if src_id > pow(2, 16) or src_id < 0:
            raise ValueError
        self.src_id = src_id

    @property
    def id_u64(self):
        return (self.src_id << 32) | (self.apid << 16) | self.seq_cnt

    def pack(self) -> bytes:
        return struct.pack("!Q", self.id_u64)

    def __str__(self):
        return (
            f"Raw u64: {self.id_u64:#016x} | APID {self.apid:#04x} | "
            f"Seq Cnt {self.seq_cnt} | SRC ID {self.src_id:#04x}"
        )

    def __repr__(self):
        return (
            f"TcSchedReqId(apid={self.apid:#04x},seq_cnt={self.seq_cnt},"
            f"src_id={self.src_id:#04x})"
        )

    @classmethod
    def build_from_tc(cls, tc: PusTelecommand) -> TcSchedReqId:
        return TcSchedReqId(tc.pus_tc_sec_header.source_id, tc.apid, tc.seq_count)



from spacepackets.seqcount import (  # noqa: F401
    ProvidesSeqCount,
    FileSeqCountProvider,
    PusFileSeqCountProvider,
    SeqCountProvider,
)

import warnings

warnings.warn(
    "the countdown module is deprecated and was moved to spacepackets.countdown",
    DeprecationWarning,
    stacklevel=2,
)



from __future__ import annotations
import logging
import os
import re
from typing import Iterable, Optional

import prompt_toolkit
from deprecated.sphinx import deprecated
from prompt_toolkit.completion import (
    CompleteEvent,
    Completer,
    Completion,
    WordCompleter,
)
from prompt_toolkit.completion.nested import NestedDict
from prompt_toolkit.document import Document
from prompt_toolkit.history import History
from prompt_toolkit.shortcuts import CompleteStyle

from tmtccmd.config.tmtc import CmdTreeNode, OpCodeEntry, TmtcDefinitionWrapper

_LOGGER = logging.getLogger(__name__)


@deprecated(
    reason="use prompt_cmd_path instead",
    version="8.0.0",
)
def prompt_service(
    tmtc_defs: TmtcDefinitionWrapper,
    compl_style: CompleteStyle = CompleteStyle.READLINE_LIKE,
) -> str:
    service_adjustment = 20
    info_adjustment = 30
    horiz_line_num = service_adjustment + info_adjustment + 3
    horiz_line = horiz_line_num * "-"
    service_ladjusted = "Service".ljust(service_adjustment)
    info_string = "Information".ljust(info_adjustment)
    tmtc_defs.sort()
    while True:
        print(f" {horiz_line}")
        print(f"|{service_ladjusted} | {info_string}|")
        print(f" {horiz_line}")
        srv_completer = build_service_word_completer(tmtc_defs)
        for service_entry in tmtc_defs.defs.items():
            try:
                adjusted_service_entry = service_entry[0].ljust(service_adjustment)
                adjusted_service_info = service_entry[1][0].ljust(info_adjustment)
                print(f"|{adjusted_service_entry} | {adjusted_service_info}|")
            except AttributeError:
                _LOGGER.warning(
                    f"Error handling service entry {service_entry[0]}. Skipping.."
                )
        print(f" {horiz_line}")
        service_string = prompt_toolkit.prompt(
            "Please select a service by specifying the key: ",
            completer=srv_completer,
            complete_style=compl_style,
        )
        if service_string in tmtc_defs.defs:
            print(f"Selected service: {service_string}")
            return service_string
        else:
            _LOGGER.warning("Invalid key, try again")


# TODO: There is an existing PR to merge this into prompt-toolkit.
# Delete this and use the package class as soon as https://github.com/prompt-toolkit/python-prompt-toolkit/pull/1815
# was merged and a new version was released.
class NestedCompleter(Completer):
    """
    Completer which wraps around several other completers, and calls any the
    one that corresponds with the first word of the input.

    By combining multiple `NestedCompleter` instances, we can achieve multiple
    hierarchical levels of autocompletion. This is useful when `WordCompleter`
    is not sufficient. The separator to trigger completion on the previously
    typed word is the Space character by default, but it is also possible
    to set a custom separator.

    If you need multiple levels, check out the `from_nested_dict` classmethod.
    """

    def __init__(
        self,
        options: dict[str, Completer | None],
        ignore_case: bool = True,
        separator: str = " ",
    ) -> None:
        self.options = options
        self.ignore_case = ignore_case
        self.separator = separator

    def __repr__(self) -> str:
        return (
            f"NestedCompleter({self.options!r}, ignore_case={self.ignore_case!r}, "
            f"separator={self.separator!r})"
        )

    @classmethod
    def from_nested_dict(
        cls, data: NestedDict, ignore_case: bool = True, separator: str = " "
    ) -> NestedCompleter:
        """
        Create a `NestedCompleter`, starting from a nested dictionary data
        structure, like this:

        .. code::

            data = {
                'show': {
                    'version': None,
                    'interfaces': None,
                    'clock': None,
                    'ip': {'interface': {'brief'}}
                },
                'exit': None
                'enable': None
            }

        The value should be `None` if there is no further completion at some
        point. If all values in the dictionary are None, it is also possible to
        use a set instead.

        Values in this data structure can be a completers as well.
        """
        options: dict[str, Completer | None] = {}
        for key, value in data.items():
            if isinstance(value, Completer):
                options[key] = value
            elif isinstance(value, dict):
                options[key] = cls.from_nested_dict(
                    data=value, ignore_case=ignore_case, separator=separator
                )
            elif isinstance(value, set):
                options[key] = cls.from_nested_dict(
                    data={item: None for item in value},
                    ignore_case=ignore_case,
                    separator=separator,
                )
            else:
                assert value is None
                options[key] = None

        return cls(options=options, ignore_case=ignore_case, separator=separator)

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:
        # Split document.
        text = document.text_before_cursor.lstrip(self.separator)
        stripped_len = len(document.text_before_cursor) - len(text)

        # If there is a separator character, check for the first term, and use a
        # subcompleter.
        if self.separator in text:
            first_term = text.split(self.separator)[0]
            completer = self.options.get(first_term)

            # If we have a sub completer, use this for the completions.
            if completer is not None:
                remaining_text = text[len(first_term) :].lstrip(self.separator)
                move_cursor = len(text) - len(remaining_text) + stripped_len

                new_document = Document(
                    remaining_text,
                    cursor_position=document.cursor_position - move_cursor,
                )

                yield from completer.get_completions(new_document, complete_event)

        # No space in the input: behave exactly like `WordCompleter`.
        else:
            completer = WordCompleter(
                list(self.options.keys()), ignore_case=self.ignore_case
            )
            yield from completer.get_completions(document, complete_event)


def prompt_cmd_path(
    cmd_def_tree: CmdTreeNode,
    history: Optional[History] = None,
    compl_style: CompleteStyle = CompleteStyle.READLINE_LIKE,
) -> str:
    compl_dict = cmd_def_tree.name_dict
    compl_dict = compl_dict.get("/")
    if compl_dict is None:
        return "/"
    nested_completer = NestedCompleter.from_nested_dict(compl_dict, separator="/")
    help_txt = (
        f"Additional commands for prompt:{os.linesep}"
        f":p[b][f][<depth>] Tree Print | :r Retry | :h Help Text | :c Cancel {os.linesep}"
        f"Auto complete is available using Tab after typing the slash character.{os.linesep}"
        f"If a command history was passed, use arrow up to access it.{os.linesep}"
        f"You can also print a subtree by typing the path and appending :p[b][p][<depth>]."
        f"{os.linesep}The b option for printouts enables brief printouts without descriptions."
        f"{os.linesep}The p option for printouts overrides hide flags to display all hidden nodes."
        f"{os.linesep}"
    )
    print(help_txt)
    while True:
        path_or_cmd = prompt_toolkit.prompt(
            message="> ",
            history=history,
            completer=nested_completer,
            complete_style=compl_style,
        )
        if ":p" in path_or_cmd:
            list_of_patterns = path_or_cmd.split(":")
            tree_to_print = cmd_def_tree
            if cmd_def_tree.contains_path(list_of_patterns[0]):
                tree_to_print = cmd_def_tree.extract_subnode(list_of_patterns[0])
                # Should never happen.
                assert tree_to_print is not None
            with_descriptions = True
            depth = None
            show_hidden_elements = False
            pattern = r"p([a-zA-Z]*)(\d*)"
            matches = re.search(pattern, list_of_patterns[1])
            if matches:
                if "b" in matches.group(1):
                    with_descriptions = False
                if "f" in matches.group(1):
                    show_hidden_elements = True
                if matches.group(2).isdigit():
                    depth = int(matches.group(2))
            print(
                tree_to_print.str_for_tree(
                    with_description=with_descriptions,
                    max_depth=depth,
                    show_hidden_elements=show_hidden_elements,
                )
            )
            continue
        elif ":h" in path_or_cmd:
            print(help_txt)
            continue
        elif ":r" in path_or_cmd:
            continue
        elif ":c" in path_or_cmd:
            break
        if not cmd_def_tree.contains_path(f"/{path_or_cmd}"):
            yes_or_no = input(
                "Command definitions tree does not contain the path. Try again? [y/n]: "
            )
            if yes_or_no in ["y", "yes", "1"]:
                continue
        break
    return f"/{path_or_cmd}"


def build_service_word_completer(
    tmtc_defs: TmtcDefinitionWrapper,
) -> WordCompleter:
    srv_list = []
    for service_entry in tmtc_defs.defs.items():
        srv_list.append(service_entry[0])
    srv_completer = WordCompleter(words=srv_list, ignore_case=True)
    return srv_completer


def prompt_op_code(
    tmtc_defs: TmtcDefinitionWrapper,
    service: str,
    compl_style: CompleteStyle = CompleteStyle.READLINE_LIKE,
) -> str:
    op_code_adjustment = 24
    info_adjustment = 56
    horz_line_num = op_code_adjustment + info_adjustment + 3
    horiz_line = horz_line_num * "-"
    op_code_info_number_str = "Operation Code (Number)".ljust(op_code_adjustment)
    op_code_info_str_str = "Operation Code (Text)".ljust(op_code_adjustment)
    info_string = "Information".ljust(info_adjustment)
    if service in tmtc_defs.defs:
        op_code_entry = tmtc_defs.op_code_entry(service)
        completer = build_op_code_word_completer(op_code_entry=op_code_entry)
    else:
        _LOGGER.warning("Service not in dictionary. Setting default operation code 0")
        return "0"
    while True:

        def print_table(otype: str, dictionary: dict):
            print(f" {horiz_line}")
            print(f"|{otype} | {info_string}|")
            print(f" {horiz_line}")
            for op_code in dictionary.items():
                adjusted_op_code_entry = op_code[0].ljust(op_code_adjustment)
                adjusted_op_code_info = op_code[1][0].ljust(info_adjustment)
                print(f"|{adjusted_op_code_entry} | {adjusted_op_code_info}|")
            print(f" {horiz_line}")

        if len(op_code_entry.op_code_dict_str_keys) > 0:
            print_table(op_code_info_str_str, op_code_entry.op_code_dict_str_keys)
        if len(op_code_entry.op_code_dict_num_keys) > 0:
            print_table(op_code_info_number_str, op_code_entry.op_code_dict_num_keys)
        op_code_string = prompt_toolkit.prompt(
            "Please select an operation code by specifying the key: ",
            completer=completer,
            complete_style=compl_style,
        )
        if (
            op_code_string in op_code_entry.op_code_dict_str_keys.keys()
            or op_code_string in op_code_entry.op_code_dict_num_keys.keys()
        ):
            print(f"Selected op code: {op_code_string}")
            return op_code_string
        else:
            _LOGGER.warning("Invalid key, try again")


def build_op_code_word_completer(op_code_entry: OpCodeEntry) -> WordCompleter:
    op_code_list = []
    for op_code_str in op_code_entry.op_code_dict_num_keys.keys():
        op_code_list.append(op_code_str)
    for op_code_str in op_code_entry.op_code_dict_str_keys.keys():
        op_code_list.append(op_code_str)
    op_code_completer = WordCompleter(words=op_code_list, ignore_case=True)
    return op_code_completer



"""
@brief      This module is used to set up the global loggers
"""

from tmtccmd.version import get_version
import logging
import os
import sys
from datetime import datetime

from deprecated.sphinx import deprecated
from colorlog import ColoredFormatter


LOG_DIR = "log"
ERROR_LOG_FILE_NAME = "tmtc_error.log"


def add_colorlog_console_logger(logger: logging.Logger, log_level: int = logging.INFO):
    """This function can be used to apply the default library console logging output format to
    a custom logger.
    """
    from colorlog import StreamHandler

    dbg_fmt = (
        "%(log_color)s%(levelname)-8s %(cyan)s%(asctime)s.%(msecs)03d "
        "[%(name)s:%(lineno)d] %(reset)s%(message)s"
    )
    custom_formatter = CustomTmtccmdFormatter(
        info_fmt=(
            "%(log_color)s%(levelname)-8s %(cyan)s%(asctime)s."
            "%(msecs)03d %(reset)s%(message)s"
        ),
        dbg_fmt=dbg_fmt,
        err_fmt=dbg_fmt,
        warn_fmt=dbg_fmt,
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    console_handler = StreamHandler(stream=sys.stdout)
    console_handler.setFormatter(custom_formatter)
    logger.addHandler(console_handler)
    logger.setLevel(log_level)


def add_error_file_logger(logger: logging.Logger):
    file_format = logging.Formatter(
        fmt=(
            "%(levelname)-8s: %(asctime)s.%(msecs)03d [%(filename)s:%(lineno)d]"
            " %(message)s"
        ),
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    # TODO: Use path relative to script dir, otherwise this craps everything
    if not os.path.exists(LOG_DIR):
        os.mkdir(LOG_DIR)
    error_file_handler = logging.FileHandler(
        filename=f"{LOG_DIR}/{ERROR_LOG_FILE_NAME}", encoding="utf-8", mode="w"
    )
    error_file_handler.setLevel(level=logging.WARNING)
    error_file_handler.setFormatter(file_format)
    logger.addHandler(error_file_handler)


@deprecated(
    version="v4.0.0a4",
    reason="use a custom application logger or tmtccmd.get_lib_logger instead.",
)
def get_console_logger() -> logging.Logger:
    return logging.getLogger(__name__)


def __setup_tmtc_console_logger(
    root_logger: logging.Logger, propagate: bool, log_level: int
):
    """Set up the tmtccmd root logger.

    :return:    Returns the instance of the global logger
    """
    # Use colorlog for now because it allows more flexibility and custom messages
    # for different levels
    add_colorlog_console_logger(logger=root_logger)
    root_logger.setLevel(level=log_level)
    root_logger.propagate = propagate


def __set_up_coloredlogs_logger(logger: logging.Logger):
    try:
        import coloredlogs

        coloredlogs.install(
            level="INFO",
            logger=logger,
            milliseconds=True,
            fmt=(
                "%(asctime)s.%(msecs)03d %(hostname)s %(name)s[%(process)d] "
                "%(levelname)s %(message)s"
            ),
            datefmt="%Y-%m-%d %H:%M:%S",
        )
    except ImportError:
        print("Please install coloredlogs package first")


# Custom formatter. Allows different strings for info, error and debug output
class CustomTmtccmdFormatter(ColoredFormatter):
    def __init__(
        self, info_fmt: str, dbg_fmt: str, err_fmt: str, warn_fmt: str, datefmt=None
    ):
        self.err_fmt = err_fmt
        self.info_fmt = info_fmt
        self.dbg_fmt = dbg_fmt
        self.warn_fmt = warn_fmt
        super().__init__(fmt="%(levelno)d: %(msg)s", datefmt=datefmt, style="%")

    def format(self, record):
        # Save the original format configured by the user
        # when the logger formatter was instantiated
        format_orig = self._style._fmt

        # Replace the original format with one customized by logging level
        if record.levelno == logging.DEBUG:
            self._style._fmt = self.dbg_fmt

        elif record.levelno == logging.INFO:
            self._style._fmt = self.info_fmt

        elif record.levelno == logging.ERROR:
            self._style._fmt = self.err_fmt

        elif record.levelno == logging.WARNING:
            self._style._fmt = self.warn_fmt

        # Call the original formatter class to do the grunt work
        result = logging.Formatter.format(self, record)

        # Restore the original format configured by the user
        self._style._fmt = format_orig

        return result


def build_log_file_name(base_name: str):
    return f"{LOG_DIR}/{base_name}"


def get_current_time_string(ms_prec: bool) -> str:
    base_fmt = "%Y-%m-%d %H:%M:%S"
    if ms_prec:
        base_fmt += ".%f"
        return datetime.now().strftime(base_fmt)[:-3]
    return datetime.now().strftime(base_fmt)



from __future__ import annotations
import enum
import logging
from pathlib import Path
from typing import Optional, Union
from datetime import datetime

from spacepackets.ecss import PusTelecommand, PusTelemetry
from tmtccmd.logging import LOG_DIR
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler
from logging import FileHandler

RAW_PUS_LOGGER_NAME = "tmtccmd_raw_pus_log"
RAW_PUS_FILE_BASE_NAME = "_".join(RAW_PUS_LOGGER_NAME.split("_")[:-1])

TMTC_LOGGER_NAME = "tmtccmd_file_log"
TMTC_FILE_BASE_NAME = "_".join(TMTC_LOGGER_NAME.split("_")[:-1])

__TMTC_LOGGER: Optional[logging.Logger] = None
__RAW_PUS_LOGGER: Optional[logging.Logger] = None


def date_suffix() -> str:
    return f"{datetime.now().date()}"


class TimedLogWhen(enum.Enum):
    PER_HOUR = "h"
    PER_MINUTE = "M"
    PER_SECOND = "S"
    PER_DAY = "D"


class RawTmtcLogBase:
    def __init__(
        self, logger: logging.Logger, log_repr: bool = True, log_raw_repr: bool = True
    ):
        self.logger = logger
        self.do_log_repr = log_repr
        self.do_log_raw_repr = log_raw_repr
        self.counter = 0

    def log_tc(self, packet: PusTelecommand):
        """Default log function which logs the Python packet representation and raw bytes"""
        prefix = self.tc_prefix(packet, self.counter)
        if self.do_log_repr:
            self.log_repr(prefix, packet)
        raw_bytes = packet.pack()
        self.__log_raw_inc_counter(prefix, raw_bytes)

    def log_tm(self, packet: PusTelemetry):
        """Default log function which logs the Python packet representation and raw bytes"""
        prefix = self.tm_prefix(packet, self.counter)
        if self.do_log_repr:
            self.log_repr(prefix, packet)
        raw_bytes = packet.pack()
        self.__log_raw_inc_counter(prefix, raw_bytes)

    def __log_raw_inc_counter(self, prefix: str, raw: bytes):
        self.log_bytes_readable(prefix, raw)
        if self.do_log_raw_repr:
            self.log_bytes_repr(prefix, raw)
        self.counter += 1

    def log_repr(self, prefix: str, packet: Union[PusTelecommand, PusTelemetry]):
        self.logger.info(f"{prefix} repr: {packet!r}")

    @staticmethod
    def tc_prefix(packet: PusTelecommand, counter: int):
        return f"tc {counter} [{packet.service}, {packet.subservice}]"

    @staticmethod
    def tm_prefix(packet: PusTelemetry, counter: int):
        return f"tm {counter} [{packet.service}, {packet.subservice}]"

    def log_bytes_readable(self, prefix: str, packet: bytes):
        self.logger.info(f"{prefix} raw readable hex: [{packet.hex(sep=',')}]")

    def log_bytes_repr(self, prefix: str, packet: bytes):
        self.logger.info(f"{prefix} raw repr: {packet!r}")


class RawTmtcTimedLogWrapper(RawTmtcLogBase):
    def __init__(
        self,
        when: TimedLogWhen,
        interval: int,
        logger: Optional[logging.Logger] = None,
        file_name: Path = Path(f"{LOG_DIR}/{RAW_PUS_FILE_BASE_NAME}"),
        suffix: Optional[str] = f"{date_suffix()}.log",
    ):
        """Create a raw TMTC timed rotating log wrapper.
        See the official Python documentation at
        https://docs.python.org/3/library/logging.handlers.html#timedrotatingfilehandler for
        more information on the input parameters

        :param when: A new log file will be created at the product of when and interval
        :param interval: A new log file will be created at the product of when and interval.
            For example, using when="H" and interval=3, a new log file will be created in three
            hour intervals
        :param file_name: Base filename of the log file
        """
        if logger is None:
            logger = logging.getLogger(RAW_PUS_LOGGER_NAME)
            logger.propagate = False
        formatter = logging.Formatter(
            fmt="%(asctime)s.%(msecs)03d: %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
        if suffix:
            file_name = f"{file_name}_{suffix}"
        handler = TimedRotatingFileHandler(
            filename=file_name, when=when.value, interval=interval
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(logging.INFO)
        self.file_name = handler.baseFilename
        super().__init__(logger)


class RawTmtcRotatingLogWrapper(RawTmtcLogBase):
    def __init__(
        self,
        max_bytes: int,
        backup_count: int,
        logger: Optional[logging.Logger] = None,
        file_name: Path = Path(f"{LOG_DIR}/{RAW_PUS_FILE_BASE_NAME}"),
        suffix: Optional[str] = f"{date_suffix()}.log",
    ):
        """Create a raw TMTC rotating log wrapper.
        See the official Python documentation at
        https://docs.python.org/3/library/logging.handlers.html#rotatingfilehandler for
        more information on the input parameters

        :param max_bytes: Maximum number of bytes per file. If backup_count is non-zero, the handler
            will create a new file up to the number of back_up count files. If the maximum backup
            count is reached, the oldest files will be deleted
        :param backup_count: If this is zero, Only max_bytes can be stored. Otherwise, a rollover
            will occur when a file reaches max_bytes and up to back_count files can be created
            this way.
        :param file_name: Base filename of the log file
        :param suffix: Suffix of the log file. Can be used to change the used log file. The default
            argument will use a date suffix, which will lead to a new unique rotating log created
            every day
        """
        if logger is None:
            logger = logging.getLogger(RAW_PUS_LOGGER_NAME)
            # We don't want any loggers up the hierarchy to receive the logs, only the file
            logger.propagate = False
        formatter = logging.Formatter(
            fmt="%(asctime)s.%(msecs)03d: %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
        if suffix:
            file_name = f"{file_name}_{suffix}"
        handler = RotatingFileHandler(
            filename=file_name,
            maxBytes=max_bytes,
            backupCount=backup_count,
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(logging.INFO)
        self.file_name = handler.baseFilename
        super().__init__(logger)


class RegularTmtcLogWrapper:
    def __init__(
        self, file_name: Optional[Path] = None, logger: Optional[logging.Logger] = None
    ):
        if logger is None:
            logger = logging.getLogger(TMTC_LOGGER_NAME)
            logger.propagate = False
        log_dir = Path(LOG_DIR)
        if not log_dir.exists():
            log_dir.mkdir()
        if file_name is None:
            self.file_name = self.get_current_tmtc_file_name()
        else:
            self.file_name = file_name
        self.logger = logger
        self.file_handler = FileHandler(self.file_name)
        formatter = logging.Formatter()
        self.file_handler.setFormatter(formatter)
        self.logger.addHandler(self.file_handler)
        self.logger.setLevel(logging.INFO)

    @classmethod
    def get_current_tmtc_file_name(cls) -> Path:
        return Path(
            f"{LOG_DIR}/{TMTC_FILE_BASE_NAME}_{datetime.now().date()}_"
            f"{datetime.now().time().strftime('%H%M%S')}.log"
        )

    def __del__(self):
        self.logger.removeHandler(self.file_handler)
        self.file_handler.close()



from __future__ import annotations
import logging
import time
from typing import Optional, Any

from PyQt6.QtCore import QRunnable, pyqtSlot, QObject, pyqtSignal
from tmtccmd.config.hook import HookBase

from tmtccmd.core import TmMode, TcMode, BackendRequest
from tmtccmd.gui.defs import LocalArgs, SharedArgs, WorkerOperationsCode
from tmtccmd.tmtc.procedure import TreeCommandingProcedure

LOGGER = logging.getLogger(__name__)


class WorkerSignalWrapper(QObject):
    finished = pyqtSignal(object)
    failure = pyqtSignal(object)
    stop = pyqtSignal(object)
    abort = pyqtSignal(object)


class FrontendWorker(QRunnable):
    def __init__(self, local_args: LocalArgs, shared_args: SharedArgs):
        super(QRunnable, self).__init__()
        self._locals = local_args
        self._shared = shared_args
        self.signals = WorkerSignalWrapper()
        self._stop_signal = False
        self._abort_signal = False
        self.signals.stop.connect(self._stop_com_if)
        self.signals.abort.connect(self._abort)

    @classmethod
    def spawn_for_opening_com_if(
        cls,
        com_if_is_switched: bool,
        com_if_key: str,
        hook: HookBase,
        shared_args: SharedArgs,
    ):
        return cls(
            LocalArgs(
                WorkerOperationsCode.OPEN_COM_IF, (com_if_is_switched, com_if_key, hook)
            ),
            shared_args,
        )

    @classmethod
    def spawn_for_cmd_path(
        cls, cmd_path: str, shared_args: SharedArgs
    ) -> FrontendWorker:
        return cls(
            LocalArgs(WorkerOperationsCode.ONE_QUEUE_MODE, cmd_path), shared_args
        )

    def __setup(self, op_code: WorkerOperationsCode) -> bool:
        if op_code == WorkerOperationsCode.OPEN_COM_IF:
            LOGGER.info("Switching COM Interface")

            assert isinstance(self._locals.op_args, tuple)
            assert isinstance(self._locals.op_args[0], bool)
            assert isinstance(self._locals.op_args[1], str)
            assert isinstance(self._locals.op_args[2], HookBase)
            # TODO: We should really pass a proper object here instead of using magic tuples..
            new_com_if = self._locals.op_args[2].get_communication_interface(
                com_if_key=self._locals.op_args[1]
            )
            # self._args.state.last_com_if = self._args.state.current_com_if
            set_success = False
            if new_com_if is not None and self._locals.op_args[0]:
                set_success = self._shared.backend.try_set_com_if(new_com_if)
            if not set_success:
                LOGGER.warning(
                    f"Could not set new communication interface {new_com_if}"
                )
            if self._shared.backend.com_if_active():
                self._finish_with_info("COM Interface is already active")
            else:
                self._shared.backend.open_com_if()
                self._finish_success()
            return False
        if op_code == WorkerOperationsCode.CLOSE_COM_IF:
            if not self._shared.backend.com_if_active():
                self._finish_with_info("COM Interface is not active")
            elif self._shared.com_if_ref_tracker.is_used():
                self._failure_with_info("Can not close COM interface which is used")
            else:
                self._shared.backend.close_com_if()
                self._finish_success()
            return False
        if op_code == WorkerOperationsCode.ONE_QUEUE_MODE:
            self._shared.com_if_ref_tracker.add_user()
            assert isinstance(self._locals.op_args, str)
            with self._shared.tc_lock:
                self._shared.backend.current_procedure = TreeCommandingProcedure(
                    self._locals.op_args
                )
                self._shared.backend.tc_mode = TcMode.ONE_QUEUE
        elif op_code == WorkerOperationsCode.LISTEN_FOR_TM:
            self._shared.com_if_ref_tracker.add_user()
            self._shared.backend.tm_mode = TmMode.LISTENER
        return True

    def __one_queue_mode_cycle(self) -> Optional[bool]:
        self._shared.tc_lock.acquire()
        self._shared.backend.tc_operation()
        self._update_backend_mode()
        state = self._shared.backend.state
        if state.request == BackendRequest.TERMINATION_NO_ERROR:
            self._shared.tc_lock.release()
            self._shared.com_if_ref_tracker.remove_user()
            self._finish_success()
            return False
        elif state.request == BackendRequest.DELAY_IDLE:
            time.sleep(1.0)
        elif state.request == BackendRequest.DELAY_CUSTOM:
            self._shared.tc_lock.release()
            if state.next_delay.total_seconds() < 0.5:
                time.sleep(state.next_delay.total_seconds())
            else:
                time.sleep(0.5)
        elif state.request == BackendRequest.CALL_NEXT:
            self._shared.tc_lock.release()

    def __listener_cycle(self) -> Optional[bool]:
        if self._stop_signal or self._abort_signal:
            self._shared.com_if_ref_tracker.remove_user()
            if not self._abort_signal:
                self._finish_success()
            return False
        else:
            # We only should run the TM operation here
            self._shared.backend.tm_operation()
            # Poll TM every 400 ms for now
            time.sleep(self._locals.op_args)

    def __loop(self, op_code: WorkerOperationsCode) -> bool:
        if op_code == WorkerOperationsCode.ONE_QUEUE_MODE:
            should_return = self.__one_queue_mode_cycle()
            if should_return is not None:
                return should_return
        elif op_code == WorkerOperationsCode.LISTEN_FOR_TM:
            should_return = self.__listener_cycle()
            if should_return is not None:
                return should_return
        elif op_code == WorkerOperationsCode.IDLE:
            return False
        else:
            # This must be a programming error
            LOGGER.error(f"Unknown worker operation code {self._locals.op_code}")
        return True

    @pyqtSlot()
    def run(self):
        op_code = self._locals.op_code
        loop_required = self.__setup(op_code)
        if loop_required:
            while True:
                if not self.__loop(op_code):
                    break

    def _finish_success(self):
        self.signals.finished.emit(None)

    def _finish_with_info(self, info: str):
        self.signals.finished.emit(info)

    def _failure_with_info(self, info: str):
        self.signals.failure.emit(info)

    def _update_backend_mode(self):
        with self._shared.state_lock:
            self._shared.backend.mode_to_req()

    def _stop_com_if(self, _args: Any):
        self._stop_signal = True

    def _abort(self, _args: Any):
        self._abort_signal = True



from __future__ import annotations

import abc
from abc import ABC
from collections import deque
from datetime import timedelta
from enum import Enum
from typing import Any, Deque, Optional, Type, cast

from spacepackets.ccsds import SpacePacket
from spacepackets.ecss import PusService, PusVerificator, check_pus_crc
from spacepackets.ecss.tc import PusTelecommand
from spacepackets.seqcount import ProvidesSeqCount

from tmtccmd.pus.s11_tc_sched import Subservice as Pus11Subservice
from tmtccmd.tmtc.procedure import TreeCommandingProcedure, TcProcedureBase


class TcQueueEntryType(Enum):
    PUS_TC = "pus-tc"
    CCSDS_TC = "ccsds-tc"
    RAW_TC = "raw-tc"
    CUSTOM = "custom"
    LOG = "log"
    WAIT = "wait"
    PACKET_DELAY = "set-delay"


class TcQueueEntryBase:
    """Generic TC queue entry abstraction. This allows filling the TC queue with custom objects"""

    def __init__(self, etype: TcQueueEntryType):
        self.etype = etype

    def is_tc(self) -> bool:
        """Check whether concrete object is an actual telecommand"""
        if (
            self.etype == TcQueueEntryType.PUS_TC
            or self.etype == TcQueueEntryType.RAW_TC
            or self.etype == TcQueueEntryType.CCSDS_TC
        ):
            return True
        return False


QueueDequeT = Deque[TcQueueEntryBase]


class PusTcEntry(TcQueueEntryBase):
    def __init__(self, pus_tc: PusTelecommand):
        super().__init__(TcQueueEntryType.PUS_TC)
        self.pus_tc = pus_tc

    def __repr__(self):
        return f"{self.__class__.__name__}({self.pus_tc!r})"


class SpacePacketEntry(TcQueueEntryBase):
    def __init__(self, space_packet: SpacePacket):
        super().__init__(TcQueueEntryType.CCSDS_TC)
        self.space_packet = space_packet

    def __repr__(self):
        return f"{self.__class__.__name__}({self.space_packet!r})"


class LogQueueEntry(TcQueueEntryBase):
    def __init__(self, log_str: str):
        super().__init__(TcQueueEntryType.LOG)
        self.log_str = log_str

    def __repr__(self):
        return f"{self.__class__.__name__}({self.log_str!r})"


class RawTcEntry(TcQueueEntryBase):
    def __init__(self, tc: bytes):
        super().__init__(TcQueueEntryType.RAW_TC)
        self.tc = tc

    def __repr__(self):
        return f"{self.__class__.__name__}({self.tc!r})"


class WaitEntry(TcQueueEntryBase):
    def __init__(self, wait_time: timedelta):
        super().__init__(TcQueueEntryType.WAIT)
        self.wait_time = wait_time

    @classmethod
    def from_millis(cls, millis: int) -> WaitEntry:
        return cls(timedelta(milliseconds=millis))

    def __repr__(self):
        return f"{self.__class__.__name__}({self.wait_time!r})"


class PacketDelayEntry(TcQueueEntryBase):
    def __init__(self, delay_time: timedelta):
        super().__init__(TcQueueEntryType.PACKET_DELAY)
        self.delay_time = delay_time

    @classmethod
    def from_millis(cls, millis: int) -> PacketDelayEntry:
        return cls(timedelta(milliseconds=millis))

    def __repr__(self):
        return f"{self.__class__.__name__}({self.delay_time!r})"


class QueueEntryHelper:
    def __init__(self, base: Optional[TcQueueEntryBase]):
        self.entry = base

    @property
    def is_tc(self) -> bool:
        assert self.entry is not None
        return self.entry.is_tc()

    @property
    def entry_type(self) -> TcQueueEntryType:
        assert self.entry is not None
        return self.entry.etype

    def __cast_internally(
        self,
        obj_type: Type[TcQueueEntryBase],
        expected_type: TcQueueEntryType,
    ) -> Any:
        assert self.entry is not None
        if self.entry.etype != expected_type:
            raise TypeError(f"Invalid object {self.entry} for type {expected_type}")
        return cast(obj_type, self.entry)

    def to_log_entry(self) -> LogQueueEntry:
        return self.__cast_internally(LogQueueEntry, TcQueueEntryType.LOG)

    def to_pus_tc_entry(self) -> PusTcEntry:
        return self.__cast_internally(PusTcEntry, TcQueueEntryType.PUS_TC)

    def to_space_packet_entry(self) -> SpacePacketEntry:
        return self.__cast_internally(SpacePacketEntry, TcQueueEntryType.CCSDS_TC)

    def to_raw_tc_entry(self) -> RawTcEntry:
        return self.__cast_internally(RawTcEntry, TcQueueEntryType.RAW_TC)

    def to_wait_entry(self) -> WaitEntry:
        return self.__cast_internally(WaitEntry, TcQueueEntryType.WAIT)

    def to_packet_delay_entry(self) -> PacketDelayEntry:
        return self.__cast_internally(PacketDelayEntry, TcQueueEntryType.PACKET_DELAY)


class QueueWrapper:
    def __init__(
        self,
        info: TcProcedureBase,
        queue: QueueDequeT,
        inter_cmd_delay: timedelta = timedelta(milliseconds=0),
    ):
        self.info = info
        self.queue = queue
        self.inter_cmd_delay = inter_cmd_delay

    @classmethod
    def empty(cls):
        return cls(TreeCommandingProcedure.empty(), deque())

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(info={self.info!r}, queue={self.queue!r}, "
            f"inter_cmd_delay={self.inter_cmd_delay!r})"
        )


class QueueHelperBase(ABC):
    def __init__(self, queue_wrapper: QueueWrapper):
        self.queue_wrapper = queue_wrapper

    def __repr__(self):
        return f"{self.__class__.__name__}(queue_wrapper={self.queue_wrapper!r})"

    def empty(self) -> bool:
        return len(self.queue_wrapper.queue) == 0

    @abc.abstractmethod
    def pre_add_cb(self, entry: TcQueueEntryBase):
        pass

    def add_log_cmd(self, print_str: str):
        self._add_entry(LogQueueEntry(print_str))

    def add_raw_tc(self, tc: bytes):
        self._add_entry(RawTcEntry(tc))

    def add_wait(self, wait_time: timedelta):
        self._add_entry(WaitEntry(wait_time))

    def add_wait_ms(self, wait_ms: int):
        self._add_entry(WaitEntry.from_millis(wait_ms))

    def add_wait_seconds(self, wait_seconds: float):
        self._add_entry(WaitEntry(timedelta(seconds=wait_seconds)))

    def add_packet_delay(self, delay: timedelta):
        self._add_entry(PacketDelayEntry(delay))

    def add_packet_delay_ms(self, delay_ms: int):
        self._add_entry(PacketDelayEntry.from_millis(delay_ms))

    def _add_entry(self, entry: TcQueueEntryBase):
        self.pre_add_cb(entry)
        self.queue_wrapper.queue.append(entry)


class DefaultPusQueueHelper(QueueHelperBase):
    """Default PUS Queue Helper which simplifies inserting PUS telecommands
    into the queue. It also provides a way to optionally stamp common PUS TC fields which would
    otherwise add boilerplate code during the packet creation process. This includes the following
    packet properties and it is also able to add the telecommand into a provided PUS verificator.

    This queue helper also has special support for PUS 11 time tagged PUS telecommands and will
    perform its core functionality for the time-tagged telecommands as well.
    """

    def __init__(
        self,
        queue_wrapper: QueueWrapper,
        tc_sched_timestamp_len: int,
        seq_cnt_provider: Optional[ProvidesSeqCount],
        pus_verificator: Optional[PusVerificator],
        default_pus_apid: Optional[int],
    ):
        """
        :param queue_wrapper: Queue Wrapper. All entries are inserted here
        :param default_pus_apid: Default APID which will be stamped onto all provided PUS TC packets
        :param seq_cnt_provider: The sequence count will be stamped onto all provided PUS TC packets
        :param pus_verificator: All provided PUS TCs will be added to this verificator
        """
        super().__init__(queue_wrapper)
        self.seq_cnt_provider = seq_cnt_provider
        self.pus_verificator = pus_verificator
        self.pus_apid = default_pus_apid
        self.tc_sched_timestamp_len = tc_sched_timestamp_len

    def pre_add_cb(self, entry: TcQueueEntryBase):
        if entry.etype == TcQueueEntryType.PUS_TC:
            pus_entry = cast(PusTcEntry, entry)
            if (
                pus_entry.pus_tc.service == PusService.S11_TC_SCHED
                and pus_entry.pus_tc.subservice == Pus11Subservice.TC_INSERT
            ):
                self._handle_time_tagged_tc(pus_entry.pus_tc)
            self._pus_packet_handler(pus_entry.pus_tc)

    def _handle_time_tagged_tc(self, pus_tc: PusTelecommand):
        new_pus_tc_app_data = bytearray()
        new_pus_tc_app_data.extend(pus_tc.app_data[: self.tc_sched_timestamp_len])
        pus_tc_raw = new_pus_tc_app_data[self.tc_sched_timestamp_len :]
        if not check_pus_crc(pus_tc_raw):
            raise ValueError(
                f"crc check on contained PUS TC with length {len(pus_tc_raw)} failed"
            )
        time_tagged_tc = PusTelecommand.unpack(pus_tc_raw)
        self._pus_packet_handler(time_tagged_tc)
        new_pus_tc_app_data.extend(time_tagged_tc.pack())
        pus_tc._app_data = new_pus_tc_app_data

    def _pus_packet_handler(self, pus_tc: PusTelecommand):
        recalc_crc = False
        if self.pus_apid is not None:
            recalc_crc = True
            pus_tc.apid = self.pus_apid
        if self.seq_cnt_provider is not None:
            recalc_crc = True
            pus_tc.seq_count = self.seq_cnt_provider.get_and_increment()
        if self.pus_verificator is not None:
            # Add TC after Sequence Count and APID stamping
            self.pus_verificator.add_tc(pus_tc)
        if recalc_crc:
            pus_tc.calc_crc()

    def add_pus_tc(self, pus_tc: PusTelecommand):
        super()._add_entry(PusTcEntry(pus_tc))

    def add_ccsds_tc(self, space_packet: SpacePacket):
        super()._add_entry(SpacePacketEntry(space_packet))



"""UDP Communication Interface"""

import logging
import select
import socket
from typing import List, Optional

from tmtccmd.com import ComInterface
from tmtccmd.com.tcpip_utils import EthAddr


_LOGGER = logging.getLogger(__name__)


class UdpClient(ComInterface):
    """Communication interface for UDP communication"""

    def __init__(
        self,
        com_if_id: str,
        send_address: EthAddr,
        recv_addr: Optional[EthAddr] = None,
    ):
        """Initialize a communication interface to send and receive UDP datagrams.

        :param send_address:
        :param recv_addr:
        """
        self.udp_socket = None
        self.com_if_id = com_if_id
        self.send_address = send_address
        self.recv_addr = recv_addr

    @property
    def id(self) -> str:
        return self.com_if_id

    def __del__(self):
        try:
            self.close()
        except IOError:
            _LOGGER.warning("Could not close UDP communication interface")

    def initialize(self, args: any = None) -> any:
        pass

    def open(self, args: any = None):
        self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # Bind is possible but should not be necessary, and introduces risk of port already
        # being used.
        # See: https://docs.microsoft.com/en-us/windows/win32/api/winsock/nf-winsock-bind
        if self.recv_addr is not None:
            _LOGGER.info(
                f"Binding UDP socket to {self.recv_addr.ip_addr} and port"
                f" {self.recv_addr.port}"
            )
            self.udp_socket.bind(self.recv_addr.to_tuple)
        # Set non-blocking because we use select
        self.udp_socket.setblocking(False)

    def is_open(self) -> bool:
        return self.udp_socket is not None

    def close(self, args: any = None) -> None:
        if self.udp_socket is not None:
            self.udp_socket.close()

    def send(self, data: bytes):
        if self.udp_socket is None:
            return
        bytes_sent = self.udp_socket.sendto(data, self.send_address.to_tuple)
        if bytes_sent != len(data):
            _LOGGER.warning("Not all bytes were sent!")

    def data_available(self, timeout: float = 0, parameters: any = 0) -> bool:
        if self.udp_socket is None:
            return False
        ready = select.select([self.udp_socket], [], [], timeout)
        if ready[0]:
            return True
        return False

    def receive(self, poll_timeout: float = 0) -> List[bytes]:
        packet_list = []
        if self.udp_socket is None:
            return packet_list
        try:
            while self.data_available(poll_timeout):
                data, sender_addr = self.udp_socket.recvfrom(4096)
                packet_list.append(bytearray(data))
            return packet_list
        except ConnectionResetError:
            _LOGGER.warning("Connection reset exception occured!")
            return []



"""Dummy Virtual Communication Interface. Currently serves to use the TMTC program without needing
external hardware or an extra socket
"""

from typing import List, Optional

from deprecated.sphinx import deprecated
from spacepackets.ccsds.time import CdsShortTimestamp
from spacepackets.ecss.pus_1_verification import (
    RequestId,
    Service1Tm,
    VerificationParams,
)
from spacepackets.ecss.pus_17_test import Service17Tm
from spacepackets.ecss.tc import PusTelecommand

from tmtccmd.com import ComInterface
from tmtccmd.config import CoreComInterfaces
from tmtccmd.pus.s1_verification import Subservice as Pus1Subservice
from tmtccmd.pus.s17_test import Subservice as Pus17Subservice
from tmtccmd.tmtc import TelemetryListT


class DummyHandler:
    def __init__(self):
        self.last_tc: Optional[PusTelecommand] = None
        self.next_telemetry_package = []
        self.current_ssc = 0
        self.reply_pending = False

    @deprecated(
        version="6.0.0",
        reason="Use insert_telecommand instead",
    )
    def pass_telecommand(self, data: bytearray):
        self.insert_telecommand(data)

    def insert_telecommand(self, data: bytes):
        self.last_tc = PusTelecommand.unpack(data)
        self.reply_pending = True
        self.generate_reply_package()

    def generate_reply_package(self):
        """Generate a reply package. Currently, this only generates a reply for a ping
        telecommand."""
        assert self.last_tc is not None
        if self.last_tc.service == 17:
            if self.last_tc.subservice == 1:
                current_time_stamp = CdsShortTimestamp.now()
                tm_packer = Service1Tm(
                    subservice=Pus1Subservice.TM_ACCEPTANCE_SUCCESS,
                    apid=self.last_tc.apid,
                    seq_count=self.current_ssc,
                    verif_params=VerificationParams(
                        req_id=RequestId(
                            self.last_tc.packet_id, self.last_tc.packet_seq_control
                        )
                    ),
                    timestamp=current_time_stamp.pack(),
                )

                self.current_ssc += 1
                tm_packet_raw = tm_packer.pack()
                self.next_telemetry_package.append(tm_packet_raw)
                tm_packer = Service1Tm(
                    subservice=Pus1Subservice.TM_START_SUCCESS,
                    apid=self.last_tc.apid,
                    seq_count=self.current_ssc,
                    verif_params=VerificationParams(
                        req_id=RequestId(
                            self.last_tc.packet_id, self.last_tc.packet_seq_control
                        )
                    ),
                    timestamp=current_time_stamp.pack(),
                )
                tm_packet_raw = tm_packer.pack()
                self.next_telemetry_package.append(tm_packet_raw)
                self.current_ssc += 1

                tm_packer = Service17Tm(
                    subservice=Pus17Subservice.TM_REPLY,
                    apid=self.last_tc.apid,
                    timestamp=current_time_stamp.pack(),
                )
                tm_packet_raw = tm_packer.pack()
                self.next_telemetry_package.append(tm_packet_raw)
                self.current_ssc += 1

                tm_packer = Service1Tm(
                    subservice=Pus1Subservice.TM_COMPLETION_SUCCESS,
                    apid=self.last_tc.apid,
                    seq_count=self.current_ssc,
                    verif_params=VerificationParams(
                        req_id=RequestId(
                            self.last_tc.packet_id, self.last_tc.packet_seq_control
                        )
                    ),
                    timestamp=current_time_stamp.pack(),
                )
                tm_packet_raw = tm_packer.pack()
                self.next_telemetry_package.append(tm_packet_raw)
                self.current_ssc += 1

    def receive_reply_package(self) -> TelemetryListT:
        if self.reply_pending:
            return_list = self.next_telemetry_package.copy()
            self.next_telemetry_package.clear()
            self.reply_pending = False
            return return_list
        else:
            return []


class DummyComIF(ComInterface):
    def __init__(self):
        self.com_if_id = CoreComInterfaces.DUMMY.value
        self.dummy_handler = DummyHandler()
        self._open = False
        self.initialized = False

    @property
    def id(self) -> str:
        return self.com_if_id

    def initialize(self, args: any = None) -> any:
        self.initialized = True

    def open(self, args: any = None) -> None:
        self._open = True

    def is_open(self) -> bool:
        return self._open

    def close(self, args: any = None) -> None:
        self._open = False

    def data_available(self, timeout: float = 0, parameters: any = 0):
        if self.dummy_handler.reply_pending:
            return True
        return False

    def receive(self, parameters: any = 0) -> List[bytes]:
        return self.dummy_handler.receive_reply_package()

    def send(self, data: bytes):
        if data is not None:
            self.dummy_handler.insert_telecommand(data)



"""Base class for Service 200 mode commanding reply handling.
"""

from __future__ import annotations

import struct
from typing import Optional

from deprecated.sphinx import deprecated
from spacepackets.ccsds.time import CcsdsTimeProvider
from spacepackets.ecss.tm import CdsShortTimestamp, PusTelemetry

from tmtccmd.pus import CustomFsfwPusService
from tmtccmd.pus.s200_fsfw_mode import Subservice
from tmtccmd.tmtc.tm_base import PusTmBase, PusTmInfoBase


class Service200FsfwReader:
    def __init__(self, tm: PusTelemetry) -> None:
        self.tm = tm
        if len(tm.source_data) < 4:
            raise ValueError("service 200 TM can not even hold object ID")
        self.object_id = tm.source_data[0:4]
        self.return_value = None
        if tm.subservice == Subservice.TM_CANT_REACH_MODE:
            self.return_value = struct.unpack("!H", tm.source_data[4:6])[0]
        self.mode = None
        self.submode = None
        if (
            tm.subservice == Subservice.TM_MODE_REPLY
            or tm.subservice == Subservice.TM_WRONG_MODE_REPLY
        ):
            self.mode = struct.unpack("!I", tm.source_data[4:8])[0]
            self.submode = tm.source_data[8]

    def contains_mode(self) -> bool:
        if self.mode is not None and self.submode is not None:
            return True
        return False

    def is_cant_reach_mode_reply(self) -> bool:
        return self.tm.subservice == Subservice.TM_CANT_REACH_MODE


@deprecated(version="7.0.0", reason="use Service200FsfwReader instead")
class Service200FsfwTm(PusTmBase, PusTmInfoBase):
    @deprecated(version="7.0.0", reason="use Service200FsfwReader instead")
    def __init__(
        self,
        subservice_id: int,
        object_id: bytes,
        time: Optional[CdsShortTimestamp],
        return_value: int = 0,
        mode: int = 0,
        submode: int = 0,
        ssc: int = 0,
        apid: int = -1,
        packet_version: int = 0b000,
        space_time_ref: int = 0b0000,
        destination_id: int = 0,
    ):
        source_data = bytearray()
        source_data.extend(object_id)
        if subservice_id == 7:
            source_data.extend(struct.pack("!H", return_value))
        elif subservice_id == 6 or subservice_id == 8:
            source_data.extend(struct.pack("!I", mode))
            source_data.append(submode)
        pus_tm = PusTelemetry(
            service=CustomFsfwPusService.SERVICE_200_MODE,
            subservice=subservice_id,
            time_provider=time,
            seq_count=ssc,
            source_data=source_data,
            apid=apid,
            packet_version=packet_version,
            space_time_ref=space_time_ref,
            destination_id=destination_id,
        )
        PusTmBase.__init__(self, pus_tm=pus_tm)
        PusTmInfoBase.__init__(self, pus_tm=pus_tm)
        self.set_packet_info("Mode Reply")

        self.is_cant_reach_mode_reply = False
        self.is_mode_reply = False
        self.object_id = object_id
        self.mode = mode
        self.submode = submode
        self.return_value = return_value
        self.__init_without_base(instance=self)

    @staticmethod
    def __init_without_base(instance: Service200FsfwTm):
        tm_data = instance.tm_data
        instance.object_id = bytes(tm_data[0:4])
        if instance.subservice == Subservice.TM_CANT_REACH_MODE:
            instance.append_packet_info(": Can't reach mode")
            instance.is_cant_reach_mode_reply = True
            instance.return_value = tm_data[4] << 8 | tm_data[5]
        elif (
            instance.subservice == Subservice.TM_MODE_REPLY
            or instance.subservice == Subservice.TM_WRONG_MODE_REPLY
        ):
            instance.is_mode_reply = True
            if instance.subservice == Subservice.TM_WRONG_MODE_REPLY:
                instance.append_packet_info(": Wrong Mode")
            elif instance.subservice == Subservice.TM_MODE_REPLY:
                instance.append_packet_info(": Mode reached")
            instance.mode = struct.unpack("!I", tm_data[4:8])[0]
            instance.submode = tm_data[8]

    @classmethod
    def __empty(cls) -> Service200FsfwTm:
        return cls(
            subservice_id=0, object_id=bytearray(4), time=CdsShortTimestamp.empty()
        )

    @classmethod
    def unpack(
        cls, raw_telemetry: bytes, time_reader: Optional[CcsdsTimeProvider]
    ) -> Service200FsfwTm:
        service_200_tm = cls.__empty()
        service_200_tm.pus_tm = PusTelemetry.unpack(
            data=raw_telemetry, time_reader=time_reader
        )
        service_200_tm.__init_without_base(instance=service_200_tm)
        return service_200_tm

    def append_telemetry_content(self, content_list: list):
        super().append_telemetry_content(content_list=content_list)
        object_id_num = struct.unpack("!I", self.object_id)[0]
        content_list.append(f"0x{object_id_num:08x}")
        if self.is_cant_reach_mode_reply:
            content_list.append(hex(self.return_value))
        elif self.is_mode_reply:
            if self.mode == 1:
                mode_string = "ON"
            elif self.mode == 2:
                mode_string = "NORMAL"
            elif self.mode == 3:
                mode_string = "RAW"
            elif self.mode == 0:
                mode_string = "OFF"
            else:
                mode_string = f"UNKNOWN ({self.mode})"
            content_list.append(mode_string)
            content_list.append(str(self.submode))

    def append_telemetry_column_headers(self, header_list: list):
        super().append_telemetry_column_headers(header_list=header_list)
        header_list.append("Object ID")
        if self.is_cant_reach_mode_reply:
            header_list.append("Return Value")
        elif self.is_mode_reply:
            header_list.append("Mode")
            header_list.append("Submode")
